"""Algorithms to compute node positions for drawing."""

import random

import networkx as nx
import numpy as np
from numpy.linalg import inv, svd

from .. import convert
from ..algorithms import max_edge_order
from ..core import SimplicialComplex

__all__ = [
    "random_layout",
    "pairwise_spring_layout",
    "barycenter_spring_layout",
    "weighted_barycenter_spring_layout",
    "pca_transform",
    "circular_layout",
    "spiral_layout",
    "barycenter_kamada_kawai_layout",
]


def random_layout(H, center=None, seed=None):
    """Position nodes uniformly at random in the unit square.

    For every node, a position is generated by choosing each of dim coordinates
    uniformly at random on the interval [0.0, 1.0).  NumPy (http://scipy.org) is
    required for this function.

    Parameters
    ----------
    H : Hypergraph or SimplicialComplex
        A position will be assigned to every node in HG.
    center : array-like, optional
        Coordinate pair around which to center the layout.
        If None (default), does not center the positions.
    seed : int, optional
        Set the random state for deterministic node layouts.
        If int, `seed` is the seed used by the random number generator,
        If None (default), random numbers are sampled from the
        numpy random number generator without initialization.

    Returns
    -------
    pos : dict
        A dictionary of positions keyed by node

    See Also
    --------
    pairwise_spring_layout
    barycenter_spring_layout
    weighted_barycenter_spring_layout

    Notes
    -----
    This function proceeds exactly as NetworkX does.

    Examples
    --------
    >>> import xgi
    >>> N = 50
    >>> ps = [0.1, 0.01]
    >>> H = xgi.random_hypergraph(N, ps)
    >>> pos = xgi.random_layout(H)

    """
    import numpy as np

    if isinstance(H, SimplicialComplex):
        H = convert.from_max_simplices(H)

    if seed is not None:
        np.random.seed(seed)

    H, center = nx.drawing.layout._process_params(H, center, 2)
    pos = np.random.rand(len(H), 2) + center
    pos = pos.astype(np.float32)
    pos = dict(zip(H, pos))

    return pos


def pairwise_spring_layout(H, seed=None, k=None, **kwargs):
    """
    Position the nodes using Fruchterman-Reingold force-directed
    algorithm using the graph projection of the hypergraph
    or the hypergraph constructed from the simplicial complex.

    Parameters
    ----------
    H : Hypergraph or SimplicialComplex
        A position will be assigned to every node in H.
    seed : int, optional
        Set the random state for deterministic node layouts.
        If int, `seed` is the seed used by the random number generator,
        If None (default), random numbers are sampled from the
        numpy random number generator without initialization.
    k : float
        The spring constant of the links. When k=None (default),
        k = 1/sqrt(N). For more information, see the documentation
        for the NetworkX spring_layout() function.
    kwargs :
        Optional arguments for the NetworkX spring_layout() function.
        See https://networkx.org/documentation/stable/reference/generated/networkx.drawing.layout.spring_layout.html

    Returns
    -------
    pos : dict
        A dictionary of positions keyed by node

    See Also
    --------
    random_layout
    barycenter_spring_layout
    weighted_barycenter_spring_layout

    Notes
    -----
    If a simplicial complex is provided the results will be based on the
    hypergraph constructed from its maximal simplices.

    Examples
    --------
    >>> import xgi
    >>> N = 50
    >>> ps = [0.1, 0.01]
    >>> H = xgi.random_hypergraph(N, ps)
    >>> pos = xgi.pairwise_spring_layout(H)
    """

    if seed is not None:
        random.seed(seed)

    if isinstance(H, SimplicialComplex):
        H = convert.from_max_simplices(H)
    G = convert.convert_to_graph(H)
    pos = nx.spring_layout(G, seed=seed, k=k, **kwargs)
    return pos


def _augmented_projection(H, weighted=False):
    """Augmented version of the the graph projection of the hypergraph
    (or simplicial complex), where phantom nodes (barycenters) are created
    for each edge composed by more than two nodes.

    Parameters
    ----------
    H : Hypergraph or SimplicialComplex
    weighted : bool (default=False)
               If True weights are assigned to all hyperedges of order d=1 (links)
               and to all connections to phantom nodes within each hyperedge of
               order d>1 to keep them together. Weights scale as the order d.


    Returns
    -------
    G : networkx.Graph
        The augmented version of the graph projection
    """
    # Creating the projected networkx Graph, I will fill it manually
    G = nx.Graph()

    # Adding real nodes
    G.add_nodes_from(list(H.nodes))

    # Adding links (edges composed by two nodes only, for which we don't use phantom nodes
    for i, j in H.edges.filterby("order", 1).members():
        if weighted == True:
            G.add_edge(i, j, weight=1)
        else:
            G.add_edge(i, j)

    # Adding phantom nodes and connections therein
    # I will start from the first int node-label available
    try:
        phantom_node_id = max([n for n in H.nodes if isinstance(n, int)]) + 1
    except ValueError:
        # The list of node-labels has no integers, so I start from 0
        phantom_node_id = 0

    # Looping over the hyperedges of different order (from triples up)
    for d in range(2, max_edge_order(H) + 1):
        # Hyperedges of order d (d=2: triplets, etc.)
        for he_id, members in H.edges.filterby("order", d).members(dtype=dict).items():
            # Adding one phantom node for each hyperedge and linking it to the nodes of
            # the hyperedge
            for n in members:
                if weighted == True:
                    G.add_edge(phantom_node_id, n, weight=d)
                else:
                    G.add_edge(phantom_node_id, n)
            phantom_node_id += 1
    return G


def barycenter_spring_layout(
    H, return_phantom_graph=False, seed=None, k=None, **kwargs
):
    """
    Position the nodes using Fruchterman-Reingold force-directed
    algorithm using an augmented version of the the graph projection
    of the hypergraph (or simplicial complex), where phantom nodes
    (barycenters) are created for each edge composed by more than two nodes.
    If a simplicial complex is provided the results will be based on the
    hypergraph constructed from its maximal simplices.

    Parameters
    ----------
    H : xgi Hypergraph or SimplicialComplex
        A position will be assigned to every node in H.
    return_phantom_graph: bool (default=False)
        If True the function returns also the augmented version of the
        the graph projection of the hypergraph (or simplicial complex).
    seed : int, RandomState instance or None  optional (default=None)
        Set the random state for deterministic node layouts.
        If int, `seed` is the seed used by the random number generator,
        If None (default), random numbers are sampled from the
        numpy random number generator without initialization.
    k : float
        The spring constant of the links. When k=None (default),
        k = 1/sqrt(N). For more information, see the documentation
        for the NetworkX spring_layout() function.
    kwargs :
        Optional arguments for the NetworkX spring_layout() function.
        See https://networkx.org/documentation/stable/reference/generated/networkx.drawing.layout.spring_layout.html


    Returns
    -------
    pos : dict
        A dictionary of positions keyed by node

    See Also
    --------
    random_layout
    pairwise_spring_layout
    weighted_barycenter_spring_layout

    Examples
    --------
    >>> import xgi
    >>> N = 50
    >>> ps = [0.1, 0.01]
    >>> H = xgi.random_hypergraph(N, ps)
    >>> pos = xgi.barycenter_spring_layout(H)
    """
    if seed is not None:
        random.seed(seed)

    if isinstance(H, SimplicialComplex):
        H = convert.from_max_simplices(H)

    G = _augmented_projection(H)

    # Creating a dictionary for the position of the nodes with the standard spring
    # layout
    pos_with_phantom_nodes = nx.spring_layout(G, seed=seed, k=k, **kwargs)

    # Retaining only the positions of the real nodes
    pos = {k: pos_with_phantom_nodes[k] for k in list(H.nodes)}

    if return_phantom_graph:
        return pos, G
    else:
        return pos


def weighted_barycenter_spring_layout(
    H, return_phantom_graph=False, seed=None, k=None, **kwargs
):
    """Position the nodes using Fruchterman-Reingold force-directed algorithm.

    This uses an augmented version of the the graph projection of the hypergraph (or
    simplicial complex), where phantom nodes (barycenters) are created for each edge of
    order d>1 (composed by more than two nodes).  Weights are assigned to all hyperedges
    of order 1 (links) and to all connections to phantom nodes within each hyperedge to
    keep them together. Weights scale as the order d.  If a simplicial complex is
    provided the results will be based on the hypergraph constructed from its maximal
    simplices.

    Parameters
    ----------
    H : Hypergraph or SimplicialComplex
        A position will be assigned to every node in H.
    return_phantom_graph: bool (default=False)
        If True the function returns also the augmented version of the
        the graph projection of the hypergraph (or simplicial complex).
    seed : int, RandomState instance or None  optional (default=None)
        Set the random state for deterministic node layouts.
        If int, `seed` is the seed used by the random number generator,
        If None (default), random numbers are sampled from the
        numpy random number generator without initialization.
    k : float
        The spring constant of the links. When k=None (default),
        k = 1/sqrt(N). For more information, see the documentation
        for the NetworkX spring_layout() function.
    kwargs :
        Optional arguments for the NetworkX spring_layout() function.
        See https://networkx.org/documentation/stable/reference/generated/networkx.drawing.layout.spring_layout.html


    Returns
    -------
    pos : dict
        A dictionary of positions keyed by node

    See Also
    --------
    random_layout
    pairwise_spring_layout
    barycenter_spring_layout

    Examples
    --------
    >>> import xgi
    >>> N = 50
    >>> ps = [0.1, 0.01]
    >>> H = xgi.random_hypergraph(N, ps)
    >>> pos = xgi.weighted_barycenter_spring_layout(H)

    """
    if seed is not None:
        random.seed(seed)

    if isinstance(H, SimplicialComplex):
        H = convert.from_max_simplices(H)

    G = _augmented_projection(H, weighted=True)

    # Creating a dictionary for node position with the standard spring layout
    pos_with_phantom_nodes = nx.spring_layout(
        G, weight="weight", seed=seed, k=k, **kwargs
    )

    # Retaining only the positions of the real nodes
    pos = {k: pos_with_phantom_nodes[k] for k in list(H.nodes)}

    if return_phantom_graph:
        return pos, G
    else:
        return pos


def pca_transform(pos, theta=0, degrees=True):
    """Transforms the positions of the nodes based on the
    principal components.

    Parameters
    ----------
    pos : dict of numpy arrays
        The output from any layout function
    theta : float, optional
        The angle between the horizontal axis and the principal axis
        measured counterclockwise, by default 0.
    degrees : bool, optional
        Whether the angle specified is in degrees (True)
        or in radians (False), by default True.

    Returns
    -------
    dict of numpy arrays
        The transformed positions.

    See Also
    --------
    random_layout
    pairwise_spring_layout
    barycenter_spring_layout
    weighted_barycenter_spring_layout
    """
    p = np.array(list(pos.values()))
    _, _, w = svd(p)

    pa = inv(w)

    if degrees:
        theta *= np.pi / 180

    r = np.array([[np.cos(theta), np.sin(theta)], [-np.sin(theta), np.cos(theta)]])
    t_p = p.dot(pa).dot(r).T
    x = t_p[0]
    y = t_p[1]

    return {n: np.array([x[i], y[i]]) for i, n in enumerate(pos.keys())}


def circular_layout(H, center=None, radius=None):
    """Position nodes on a circle.

    Parameters
    ----------
    H : Hypergraph or SimplicialComplex
        A position will be assigned to every node in H.
    center : array-like or None
        Coordinate pair around which to center the layout.
        If None set to [0,0]
    radius : float or None (default=None)
        Radius of the circle on which to draw the nodes,
        if None set to 1.0.

    Returns
    -------
    pos : dict
        A dictionary of positions keyed by node
    """
    if center is None:
        center = [0, 0]

    if radius is None:
        radius = 1.0

    if H.num_nodes == 0:
        pos = {}
    elif H.num_nodes == 1:
        pos = {list(H.nodes)[0]: center}
    else:
        theta = np.linspace(0, 1, len(H) + 1)[:-1] * 2 * np.pi
        pos = np.column_stack(
            [radius * np.cos(theta) + center[0], radius * np.sin(theta) + center[1]]
        )
        pos = dict(zip(list(H.nodes), pos))

    return pos


def spiral_layout(H, center=None, resolution=0.35, equidistant=False):
    """Position nodes in a spiral layout.

    Parameters
    ----------
    H : Hypergraph or SimplicialComplex
        A position will be assigned to every node in H.
    center : array-like or None
        Coordinate pair around which to center the layout.
        If None set to [0,0]
    resolution : float, default=0.35
        The compactness of the spiral layout returned.
        Lower values result in more compressed spiral layouts.
    equidistant : bool, default=False
        If True, nodes will be positioned equidistant from each other
        by decreasing angle further from center.
        If False, nodes will be positioned at equal angles
        from each other by increasing separation further from center.

    Returns
    -------
    pos : dict
        A dictionary of positions keyed by node
    """
    if center is None:
        center = [0, 0]

    if H.num_nodes == 0:
        pos = {}
        return pos
    elif H.num_nodes == 1:
        pos = {list(H.nodes)[0]: center}
        return pos

    pos = []
    if equidistant:
        chord = 1
        step = 0.5
        theta = resolution
        theta += chord / (step * theta)
        for _ in range(len(H)):
            r = step * theta
            theta += chord / r
            pos.append([np.cos(theta) * r + center[0], np.sin(theta) * r + center[1]])
    else:
        dist = np.arange(len(H), dtype=float)
        angle = resolution * dist
        pos = np.transpose(dist * np.array([np.cos(angle), np.sin(angle)]))

    pos = dict(zip(list(H.nodes), pos))

    return pos


def barycenter_kamada_kawai_layout(H, return_phantom_graph=False, **kwargs):
    """Position nodes using Kamada-Kawai path-length cost-function
    using an augmented version of the the graph projection
    of the hypergraph (or simplicial complex), where phantom nodes
    (barycenters) are created for each edge composed by more than two nodes.
    If a simplicial complex is provided the results will be based on the
    hypergraph constructed from its maximal simplices.

    Parameters
    ----------
    H : xgi Hypergraph or SimplicialComplex
        A position will be assigned to every node in H.
    return_phantom_graph: bool (default=False)
        If True the function returns also the augmented version of the
        the graph projection of the hypergraph (or simplicial complex).
    kwargs :
        Optional arguments for the NetworkX spring_layout() function.
        See https://networkx.org/documentation/stable/reference/generated/networkx.drawing.layout.kamada_kawai_layout.html

    Returns
    -------
    pos : dict
        A dictionary of positions keyed by node
    """
    if isinstance(H, SimplicialComplex):
        H = convert.from_max_simplices(H)

    G = _augmented_projection(H)

    # Creating a dictionary for the position of the nodes with the standard spring layout
    pos_with_phantom_nodes = nx.kamada_kawai_layout(G, **kwargs)

    # Retaining only the positions of the real nodes
    pos = {k: pos_with_phantom_nodes[k] for k in list(H.nodes)}

    if return_phantom_graph:
        return pos, G
    else:
        return pos



"""Draw hypergraphs and simplicial complexes with matplotlib."""

from collections.abc import Iterable

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LinearSegmentedColormap, ListedColormap
from numpy import ndarray

from ..exception import XGIError
from ..stats import IDStat
from .layout import barycenter_spring_layout


def _draw_init(H, ax, pos):
    """Initializes the axis and node positions

    Parameters
    ----------
    H : Hypergraph
        The hypergraph to draw
    ax : Matplotlib axis
        The axis on which to plot
    pos : dict of lists
        The output of the layout functions.
        If None, uses the barycenter spring layout.

    Returns
    -------
    ax, pos
        ax: axis updated with the XGI format
        pos: dict of lists with the node positions
    """

    if pos is None:
        pos = barycenter_spring_layout(H)

    if ax is None:
        ax = plt.gca()

    ax.get_xaxis().set_ticks([])
    ax.get_yaxis().set_ticks([])
    ax.axis("off")

    return ax, pos


def _update_lims(pos, ax):
    """Update Axis limits based on node positions

    Parameters
    ----------
    pos : dict of lists
        The output of the layout functions.
    ax : Matplotlib axis
        The axis on which to plot

    Returns
    -------
    None

    """

    # compute axis limits
    pos_arr = np.asarray([[x, y] for n, (x, y) in pos.items()])

    maxx, maxy = np.max(pos_arr, axis=0)
    minx, miny = np.min(pos_arr, axis=0)
    w = maxx - minx
    h = maxy - miny

    # update view after drawing
    padx, pady = 0.05 * w, 0.05 * h
    corners = (minx - padx, miny - pady), (maxx + padx, maxy + pady)
    ax.update_datalim(corners)
    ax.autoscale_view()


def _draw_arg_to_arr(arg):
    """Convert drawing arguments to a matplotlib-compliant format.

    IDStat, dict, and list are converted to ndarray.
    Scalar values are untouched.

    Parameters
    ----------
    arg: int, float, dict, iterable, or NodeStat/EdgeStat
        Attributes for drawing parameter. Scalars are ignored.

    Returns
    -------
    arg : ndarray
        Drawing argument in matplotlib-comliant form (scalar or array)
    """
    if isinstance(arg, IDStat):
        arg = arg.asnumpy()
    elif isinstance(arg, dict):
        values = list(arg.values())
        arg = np.array(values)
    elif isinstance(arg, list):
        arg = np.array(arg)

    return arg


def _interp_draw_arg(arg, min_val, max_val):
    """Linearly interpolate drawing arguments between min/max values

    Parameters
    ----------
    arg: arr-like
        Attributes for drawing parameter.

    Returns
    -------
    vals : ndarray
        Drawing argument interpolated
    """

    vals = np.interp(arg, [min(arg), max(arg)], [min_val, max_val])

    return vals


def _scalar_arg_to_dict(scalar_arg, ids, min_val, max_val):
    """Map different types of arguments for drawing style to a dict with scalar values.

    Parameters
    ----------
    scalar_arg : int, float, dict, iterable, or NodeStat/EdgeStat
        Attributes for drawing parameter.
    ids : NodeView or EdgeView
        This is the node or edge IDs that attributes get mapped to.
    min_val : int or float
        The minimum value of the drawing parameter.
    max_val : int or float
        The maximum value of the drawing parameter.

    Returns
    -------
    dict
        An ID: scalar dictionary.

    Raises
    ------
    TypeError
        If a int, float, list, dict, or NodeStat/EdgeStat is not passed.
    """
    if isinstance(scalar_arg, str):
        raise TypeError(
            "Argument must be int, float, dict, iterable, "
            f"or NodeStat/EdgeStat. Received {type(scalar_arg)}"
        )

    # Single argument
    if isinstance(scalar_arg, (int, float)):
        return {id: scalar_arg for id in ids}

    # IDStat
    if isinstance(scalar_arg, IDStat):
        vals = np.interp(
            scalar_arg.asnumpy(),
            [scalar_arg.min(), scalar_arg.max()],
            [min_val, max_val],
        )
        return dict(zip(ids, vals))

    # Iterables of floats or ints
    if isinstance(scalar_arg, Iterable):
        if isinstance(scalar_arg, dict):
            try:
                return {id: float(scalar_arg[id]) for id in scalar_arg if id in ids}
            except ValueError as e:
                raise TypeError(
                    "The input dict must have values that can be cast to floats."
                )

        elif isinstance(scalar_arg, (list, ndarray)):
            try:
                return {id: float(scalar_arg[idx]) for idx, id in enumerate(ids)}
            except ValueError as e:
                raise TypeError(
                    "The input list or array must have values that can be cast to floats."
                )
        else:
            raise TypeError(
                "Argument must be an dict, list, or numpy array of floats or ints."
            )

    raise TypeError(
        "Argument must be int, float, dict, iterable, "
        f"or NodeStat/EdgeStat. Received {type(scalar_arg)}"
    )


def _color_arg_to_dict(color_arg, ids, cmap):
    """Map different types of arguments for drawing style to a dict with color values.

    Parameters
    ----------
    color_arg : Several formats are accepted:

        Single color values

        * str
        * 3- or 4-tuple

        Iterable of colors (each color specified as above)

        * numpy array
        * list
        * dict {id: color} pairs

        Iterable of numerical values (floats or ints)

        * list
        * dict
        * numpy array

        Stats

        * NodeStat
        * EdgeStat

        Attributes for drawing parameter.
    ids : NodeView or EdgeView
        This is the node or edge IDs that attributes get mapped to.
    cmap : ListedColormap or LinearSegmentedColormap
        colormap to use for NodeStat/EdgeStat.

    Returns
    -------
    dict
        An ID: color dictionary.

    Raises
    ------
    TypeError
        If a string, dict, iterable, or NodeStat/EdgeStat is not passed.

    Notes
    -----
    For the iterable of values, we do not accept tuples,
    because there is the potential for ambiguity.
    """

    # single argument. Must be a string or a tuple of floats
    if isinstance(color_arg, str) or (
        isinstance(color_arg, tuple) and isinstance(color_arg[0], float)
    ):
        return {id: color_arg for id in ids}

    # Iterables of colors. The values of these iterables must strings or tuples. As of now,
    # there is not a check to verify that the tuples contain floats.
    if isinstance(color_arg, Iterable):
        if isinstance(color_arg, dict) and isinstance(
            next(iter(color_arg.values())), (str, tuple, ndarray)
        ):
            return {id: color_arg[id] for id in color_arg if id in ids}
        if isinstance(color_arg, (list, ndarray)) and isinstance(
            color_arg[0], (str, tuple, ndarray)
        ):
            return {id: color_arg[idx] for idx, id in enumerate(ids)}

    # Stats or iterable of values
    if isinstance(color_arg, (Iterable, IDStat)):
        # set max and min of interpolation based on color map
        if isinstance(cmap, ListedColormap):
            minval = 0
            maxval = cmap.N
        elif isinstance(cmap, LinearSegmentedColormap):
            minval = 0.1
            maxval = 0.9
        else:
            raise XGIError("Invalid colormap!")

        # handle the case of IDStat vs iterables
        if isinstance(color_arg, IDStat):
            vals = np.interp(
                color_arg.asnumpy(),
                [color_arg.min(), color_arg.max()],
                [minval, maxval],
            )
            return {
                id: np.array(cmap(vals[i])).reshape(1, -1) for i, id in enumerate(ids)
            }

        elif isinstance(color_arg, Iterable):
            if isinstance(color_arg, dict) and isinstance(
                next(iter(color_arg.values())), (int, float)
            ):
                v = list(color_arg.values())
                vals = np.interp(v, [np.min(v), np.max(v)], [minval, maxval])
                # because we have ids, we can't just assume that the keys of arg correspond to
                # the ids.
                return {
                    id: np.array(cmap(v)).reshape(1, -1)
                    for v, id in zip(vals, color_arg.keys())
                    if id in ids
                }

            if isinstance(color_arg, (list, ndarray)) and isinstance(
                color_arg[0], (int, float)
            ):
                vals = np.interp(
                    color_arg, [np.min(color_arg), np.max(color_arg)], [minval, maxval]
                )
                return {
                    id: np.array(cmap(vals[i])).reshape(1, -1)
                    for i, id in enumerate(ids)
                }
            else:
                raise TypeError(
                    "Argument must be an dict, list, or numpy array of floats."
                )

    raise TypeError(
        "Argument must be str, dict, iterable, or "
        f"NodeStat/EdgeStat. Received {type(color_arg)}"
    )


def _CCW_sort(p):
    """
    Sort the input 2D points counterclockwise.
    """
    p = np.array(p)
    mean = np.mean(p, axis=0)
    d = p - mean
    s = np.arctan2(d[:, 0], d[:, 1])
    return p[np.argsort(s), :]



"""Generate random uniform hypergraphs."""
import itertools
import operator
import random
import warnings
from functools import reduce

import numpy as np

from ..exception import XGIError
from .classic import empty_hypergraph

__all__ = [
    "uniform_hypergraph_configuration_model",
    "uniform_HSBM",
    "uniform_HPPM",
    "uniform_erdos_renyi_hypergraph",
]


def uniform_hypergraph_configuration_model(k, m, seed=None):
    """
    A function to generate an m-uniform configuration model

    Parameters
    ----------
    k : dictionary
        This is a dictionary where the keys are node ids
        and the values are node degrees.
    m : int
        specifies the hyperedge size
    seed : integer or None (default)
        The seed for the random number generator

    Returns
    -------
    Hypergraph object
        The generated hypergraph

    Warns
    -----
    warnings.warn
        If the sums of the degrees are not divisible by m, the
        algorithm still runs, but raises a warning and adds an
        additional connection to random nodes to satisfy this
        condition.

    Notes
    -----
    This algorithm normally creates multi-edges and loopy hyperedges.
    We remove the loopy hyperedges.

    References
    ----------
    "The effect of heterogeneity on hypergraph contagion models"
    by Nicholas W. Landry and Juan G. Restrepo
    https://doi.org/10.1063/5.0020034


    Example
    -------
    >>> import xgi
    >>> import random
    >>> n = 1000
    >>> m = 3
    >>> k = {1: 1, 2: 2, 3: 3, 4: 3}
    >>> H = xgi.uniform_hypergraph_configuration_model(k, m)

    """
    if seed is not None:
        random.seed(seed)

    # Making sure we have the right number of stubs
    remainder = sum(k.values()) % m
    if remainder != 0:
        warnings.warn(
            "This degree sequence is not realizable. "
            "Increasing the degree of random nodes so that it is."
        )
        random_ids = random.sample(list(k.keys()), int(round(m - remainder)))
        for id in random_ids:
            k[id] = k[id] + 1

    stubs = []
    # Creating the list to index through
    for id in k:
        stubs.extend([id] * int(k[id]))

    H = empty_hypergraph()
    H.add_nodes_from(k.keys())

    while len(stubs) != 0:
        u = random.sample(range(len(stubs)), m)
        edge = set()
        for index in u:
            edge.add(stubs[index])
        if len(edge) == m:
            H.add_edge(edge)

        for index in sorted(u, reverse=True):
            del stubs[index]

    return H


def uniform_HSBM(n, m, p, sizes, seed=None):
    """Create a uniform hypergraph stochastic block model (HSBM).

    Parameters
    ----------
    n : int
        The number of nodes
    m : int
        The hyperedge size
    p : m-dimensional numpy array
        tensor of probabilities between communities
    sizes : list or 1D numpy array
        The sizes of the community blocks in order
    seed : integer or None (default)
        The seed for the random number generator

    Returns
    -------
    Hypergraph
        The constructed SBM hypergraph

    Raises
    ------
    XGIError
        - If the length of sizes and p do not match.
        - If p is not a tensor with every dimension equal
        - If p is not m-dimensional
        - If the entries of p are not in the range [0, 1]
        - If the sum of the vector of sizes does not equal the number of nodes.
    Exception
        If there is an integer overflow error

    See Also
    --------
    uniform_HPPM

    References
    ----------
    Nicholas W. Landry and Juan G. Restrepo.
    "Polarization in hypergraphs with community structure."
    Preprint, 2023. https://doi.org/10.48550/arXiv.2302.13967
    """

    # Check if dimensions match
    if len(sizes) != np.size(p, axis=0):
        raise XGIError("'sizes' and 'p' do not match.")
    if len(np.shape(p)) != m:
        raise XGIError("The dimension of p does not match m")
    # Check that p has the same length over every dimension.
    if len(set(np.shape(p))) != 1:
        raise XGIError("'p' must be a square tensor.")
    if np.max(p) > 1 or np.min(p) < 0:
        raise XGIError("Entries of 'p' not in [0,1].")
    if np.sum(sizes) != n:
        raise XGIError("Sum of sizes does not match n")

    if seed is not None:
        np.random.seed(seed)

    node_labels = range(n)
    H = empty_hypergraph()
    H.add_nodes_from(node_labels)

    block_range = range(len(sizes))
    # Split node labels in a partition (list of sets).
    size_cumsum = [sum(sizes[0:x]) for x in range(0, len(sizes) + 1)]
    partition = [
        list(node_labels[size_cumsum[x] : size_cumsum[x + 1]])
        for x in range(0, len(size_cumsum) - 1)
    ]

    for block in itertools.product(block_range, repeat=m):
        if p[block] == 1:  # Test edges cases p_ij = 0 or 1
            edges = itertools.product((partition[i] for i in block_range))
            for e in edges:
                H.add_edge(e)
        elif p[block] > 0:
            partition_sizes = [len(partition[i]) for i in block]
            max_index = reduce(operator.mul, partition_sizes, 1)
            if max_index < 0:
                raise Exception("Index overflow error!")
            index = np.random.geometric(p[block]) - 1

            while index < max_index:
                indices = _index_to_edge_partition(index, partition_sizes, m)
                e = {partition[block[i]][indices[i]] for i in range(m)}
                if len(e) == m:
                    H.add_edge(e)
                index += np.random.geometric(p[block])
    return H


def uniform_HPPM(n, m, k, epsilon, rho=0.5, seed=None):
    """Construct the m-uniform hypergraph planted partition model (m-HPPM)

    Parameters
    ----------
    n : int > 0
        Number of nodes
    m : int > 0
        Hyperedge size
    k : float > 0
        Mean degree
    epsilon : float > 0
        Imbalance parameter
    rho : float between 0 and 1, optional
        The fraction of nodes in community 1, default 0.5
    seed : integer or None (default)
        The seed for the random number generator

    Returns
    -------
    Hypergraph
        The constructed m-HPPM hypergraph.

    Raises
    ------
    XGIError
        - If rho is not between 0 and 1
        - If the mean degree is negative.
        - If epsilon is not between 0 and 1

    See Also
    --------
    uniform_HSBM

    References
    ----------
    Nicholas W. Landry and Juan G. Restrepo.
    "Polarization in hypergraphs with community structure."
    Preprint, 2023. https://doi.org/10.48550/arXiv.2302.13967
    """

    if rho < 0 or rho > 1:
        raise XGIError("The value of rho must be between 0 and 1")
    if k < 0:
        raise XGIError("The mean degree must be non-negative")
    if epsilon < 0 or epsilon > 1:
        raise XGIError("epsilon must be between 0 and 1")

    sizes = [int(rho * n), n - int(rho * n)]

    p = k / (m * n ** (m - 1))
    # ratio of inter- to intra-community edges
    q = rho**m + (1 - rho) ** m
    r = 1 / q - 1
    p_in = (1 + r * epsilon) * p
    p_out = (1 - epsilon) * p

    p = p_out * np.ones([2] * m)
    p[tuple([0] * m)] = p_in
    p[tuple([1] * m)] = p_in

    return uniform_HSBM(n, m, p, sizes, seed=seed)


def uniform_erdos_renyi_hypergraph(n, m, p, p_type="degree", seed=None):
    """Generate an m-uniform Erdős–Rényi hypergraph

    This creates a hypergraph with `n` nodes where
    hyperedges of size `m` are created at random to
    obtain a mean degree of `k`.

    Parameters
    ----------
    n : int > 0
        Number of nodes
    m : int > 0
        Hyperedge size
    p : float or int > 0
        Mean expected degree if p_type="degree" and
        probability of an m-hyperedge if p_type="prob"
    p_type : str
        "degree" or "prob", by default "degree"
    seed : integer or None (default)
        The seed for the random number generator

    Returns
    -------
    Hypergraph
        The Erdos Renyi hypergraph


    See Also
    --------
    random_hypergraph
    """
    if seed is not None:
        np.random.seed(seed)

    node_labels = range(n)
    H = empty_hypergraph()
    H.add_nodes_from(node_labels)

    if p_type == "degree":
        q = p / (m * n ** (m - 1))  # wiring probability
    elif p_type == "prob":
        q = p
    else:
        raise XGIError("Invalid p_type!")

    if q > 1 or q < 0:
        raise XGIError("Probability not in [0,1].")

    index = np.random.geometric(q) - 1  # -1 b/c zero indexing
    max_index = n**m
    while index < max_index:
        e = set(_index_to_edge(index, n, m))
        if len(e) == m:
            H.add_edge(e)
        index += np.random.geometric(q)
    return H


def _index_to_edge(index, n, m):
    """Generate a hyperedge given an index in the list of possible edges.

    Parameters
    ----------
    index : int > 0
        The index of the hyperedge in the list of all possible hyperedges.
    n : int > 0
        The number of nodes
    m : int > 0
        The hyperedge size.

    Returns
    -------
    list
        The reconstructed hyperedge

    See Also
    --------
    _index_to_edge_partition

    References
    ----------
    https://stackoverflow.com/questions/53834707/element-at-index-in-itertools-product
    """
    return [(index // (n**r) % n) for r in range(m - 1, -1, -1)]


def _index_to_edge_partition(index, partition_sizes, m):
    """Generate a hyperedge given an index in the list of possible edges
    and a partition of community labels.

    Parameters
    ----------
    index : int > 0
        The index of the hyperedge in the list of all possible hyperedges.
    n : int > 0
        The number of nodes
    m : int > 0
        The hyperedge size.

    Returns
    -------
    list
        The reconstructed hyperedge

    See Also
    --------
    _index_to_edge

    """
    try:
        return [
            int(index // np.prod(partition_sizes[r + 1 :]) % partition_sizes[r])
            for r in range(m)
        ]
    except KeyError:
        raise Exception("Invalid parameters")



"""Method for projecting a hypergraph to a graph."""

import networkx as nx

from ..linalg import adjacency_matrix

__all__ = ["convert_to_graph"]


def convert_to_graph(H):
    """Graph projection (1-skeleton) of the hypergraph H.
    Weights are not considered.

    Parameters
    ----------
    H : Hypergraph object
        The hypergraph of interest

    Returns
    -------
    G : networkx.Graph
        The graph projection
    """

    A = adjacency_matrix(H)  # This is unweighted by design
    G = nx.from_scipy_sparse_array(A)
    G = nx.relabel_nodes(G, {i: node for i, node in enumerate(H.nodes)})
    return G



"""Generators for some lattice hypergraphs.

All the functions in this module return a Hypergraph class (i.e. a simple, undirected
hypergraph).

"""

from warnings import warn

from ..exception import XGIError

__all__ = [
    "ring_lattice",
]


def ring_lattice(n, d, k, l):
    """A ring lattice hypergraph.

    A d-uniform hypergraph on n nodes where each node is part of k edges and the
    overlap between consecutive edges is d-l.

    Parameters
    ----------
    n : int
        Number of nodes
    d : int
        Edge size
    k : int
        Number of edges of which a node is a part. Should be a multiple of 2.
    l : int
        Overlap between edges

    Returns
    -------
    Hypergraph
        The generated hypergraph

    Raises
    ------
    XGIError
        If k is negative.

    Notes
    -----
    ring_lattice(n, 2, k, 0) is a ring lattice graph where each node has k//2 edges on
    either side.

    """
    from ..core import Hypergraph

    if k < 0:
        raise XGIError("Invalid k value!")

    if k < 2:
        warn("This creates a completely disconnected hypergraph!")

    if k % 2 != 0:
        warn("k is not divisible by 2")

    edges = [
        [node] + [(start + l + i) % n for i in range(d - 1)]
        for node in range(n)
        for start in range(node + 1, node + k // 2 + 1)
    ]
    H = Hypergraph(edges)
    H.add_nodes_from(range(n))
    return H



from . import (
    classic,
    uniform,
    random,
    simple,
    lattice,
    simplicial_complexes,
    randomizing,
)
from .classic import *
from .random import *
from .uniform import *
from .simple import *
from .lattice import *
from .simplicial_complexes import *
from .randomizing import *



from . import (
    bipartite_edges,
    bipartite_graph,
    graph,
    higher_order_network,
    hyperedges,
    hypergraph_dict,
    incidence,
    line_graph,
    encapsulation_dag,
    pandas,
    simplex,
)
from .bipartite_edges import *
from .bipartite_graph import *
from .graph import *
from .higher_order_network import *
from .hyperedges import *
from .hypergraph_dict import *
from .incidence import *
from .line_graph import *
from .encapsulation_dag import *
from .pandas import *
from .simplex import *



"""Algorithms for computing the centralities of nodes (and edges) in a hypergraph."""

from warnings import warn

import networkx as nx
import numpy as np
from numpy.linalg import norm
from scipy.sparse.linalg import eigsh

from ..convert import to_line_graph
from ..exception import XGIError
from ..linalg import clique_motif_matrix, incidence_matrix
from ..utils import convert_labels_to_integers
from .properties import is_uniform

__all__ = [
    "clique_eigenvector_centrality",
    "h_eigenvector_centrality",
    "node_edge_centrality",
    "line_vector_centrality",
    "katz_centrality",
]


def clique_eigenvector_centrality(H, tol=1e-6):
    """Compute the clique motif eigenvector centrality of a hypergraph.

    Parameters
    ----------
    H : Hypergraph
        The hypergraph of interest.
    tol : float, optional
        The tolerance when computing the eigenvector. By default, 1e-6.

    Returns
    -------
    dict
        Centrality, where keys are node IDs and values are centralities. The
        centralities are 1-normalized.

    See Also
    --------
    h_eigenvector_centrality

    References
    ----------
    Three Hypergraph Eigenvector Centralities,
    Austin R. Benson,
    https://doi.org/10.1137/18M1203031
    """
    from ..algorithms import is_connected

    # if there aren't any nodes, return an empty dict
    if H.num_nodes == 0:
        return dict()
    # if the hypergraph is not connected,
    # this metric doesn't make sense and should return nan.
    if not is_connected(H):
        return {n: np.nan for n in H.nodes}
    W, node_dict = clique_motif_matrix(H, index=True)
    _, v = eigsh(W.asfptype(), k=1, which="LM", tol=tol)

    # multiply by the sign to try and enforce positivity
    v = np.sign(v[0]) * v / norm(v, 1)
    return {node_dict[n]: v[n].item() for n in node_dict}


def h_eigenvector_centrality(H, max_iter=100, tol=1e-6):
    """Compute the H-eigenvector centrality of a uniform hypergraph.

    Parameters
    ----------
    H : Hypergraph
        The hypergraph of interest.
    max_iter : int, optional
        The maximum number of iterations before the algorithm terminates.
        By default, 100.
    tol : float > 0, optional
        The desired L2 error in the centrality vector. By default, 1e-6.

    Returns
    -------
    dict
        Centrality, where keys are node IDs and values are centralities. The
        centralities are 1-normalized.

    Raises
    ------
    XGIError
        If the hypergraph is not uniform.

    See Also
    --------
    clique_eigenvector_centrality

    References
    ----------
    Three Hypergraph Eigenvector Centralities,
    Austin R. Benson,
    https://doi.org/10.1137/18M1203031
    """
    from ..algorithms import is_connected

    # if there aren't any nodes, return an empty dict
    if H.num_nodes == 0:
        return dict()
    # if the hypergraph is not connected,
    # this metric doesn't make sense and should return nan.
    if not is_connected(H):
        return {n: np.nan for n in H.nodes}

    m = is_uniform(H)
    if not m:
        raise XGIError("This method is not defined for non-uniform hypergraphs.")

    new_H = convert_labels_to_integers(H, "old-label")

    f = lambda v, m: np.power(v, 1.0 / m)  # noqa: E731
    g = lambda v, x: np.prod(v[list(x)])  # noqa: E731

    x = np.random.uniform(size=(new_H.num_nodes))
    x = x / norm(x, 1)

    for iter in range(max_iter):
        new_x = apply(new_H, x, g)
        new_x = f(new_x, m)
        # multiply by the sign to try and enforce positivity
        new_x = np.sign(new_x[0]) * new_x / norm(new_x, 1)
        if norm(x - new_x) <= tol:
            break
        x = new_x.copy()
    else:
        warn("Iteration did not converge!")
    return {new_H.nodes[n]["old-label"]: c for n, c in zip(new_H.nodes, new_x)}


def apply(H, x, g=lambda v, e: np.sum(v[list(e)])):
    """Apply a vector to the hypergraph given a function.

    Parameters
    ----------
    H : Hypergraph
        Hypergraph of interest.
    x : 1D numpy array
        1D vector
    g : lambda function, optional
        function to apply. By default, sum.

    Returns
    -------
    1D numpy array
        vector post application
    """
    new_x = np.zeros(H.num_nodes)
    for edge in H.edges.members():
        edge = list(edge)
        # ordered permutations
        for shift in range(len(edge)):
            new_x[edge[shift]] += g(x, edge[shift + 1 :] + edge[:shift])
    return new_x


def node_edge_centrality(
    H,
    f=lambda x: np.power(x, 2),
    g=lambda x: np.power(x, 0.5),
    phi=lambda x: np.power(x, 2),
    psi=lambda x: np.power(x, 0.5),
    max_iter=100,
    tol=1e-6,
):
    """Computes the node and edge centralities

    Parameters
    ----------
    H : Hypergraph
        The hypergraph of interest
    f : lambda function, optional
        The function f as described in Tudisco and Higham.
        Must accept a numpy array. By default, :math:`f(x) = x^2`.
    g : lambda function, optional
        The function g as described in Tudisco and Higham.
        Must accept a numpy array. By default, :math`g(x) = \sqrt{x}`.
    phi : lambda function, optional
        The function phi as described in Tudisco and Higham.
        Must accept a numpy array. By default :math:`\phi(x) = x^2`.
    psi : lambda function, optional
        The function psi as described in Tudisco and Higham.
        Must accept a numpy array. By default: :math:`\psi(x) = \sqrt{x}`.
    max_iter : int, optional
        Number of iterations at which the algorithm terminates
        if convergence is not reached. By default, 100.
    tol : float > 0, optional
        The total allowable error in the node and edge centralities.
        By default, 1e-6.

    Returns
    -------
    dict, dict
        The node centrality where keys are node IDs and values are associated
        centralities and the edge centrality where keys are the edge IDs and
        values are associated centralities. The centralities of both the nodes
        and edges are 1-normalized.

    Notes
    -----
    In the paper from which this was taken, it is more general in that it includes
    general functions for both nodes and edges, nodes and edges may be weighted,
    and one can choose different norms for normalization.

    References
    ----------
    Node and edge nonlinear eigenvector centrality for hypergraphs,
    Francesco Tudisco & Desmond J. Higham,
    https://doi.org/10.1038/s42005-021-00704-2
    """
    from ..algorithms import is_connected

    # if there aren't any nodes or edges, return an empty dict
    if H.num_nodes == 0 or H.num_edges == 0 or not is_connected(H):
        return {n: np.nan for n in H.nodes}, {e: np.nan for e in H.edges}
    # if the hypergraph is not connected,
    # this metric doesn't make sense and should return nan.
    # if not is_connected(H):
    #     return {n: np.nan for n in H.nodes}, {e: np.nan for e in H.edges}

    n = H.num_nodes
    m = H.num_edges
    x = np.ones(n) / n
    y = np.ones(m) / m

    I, node_dict, edge_dict = incidence_matrix(H, index=True)

    check = np.inf

    for iter in range(max_iter):
        u = np.multiply(x, g(I @ f(y)))
        v = np.multiply(y, psi(I.T @ phi(x)))
        # multiply by the sign to try and enforce positivity
        new_x = np.sign(u[0]) * u / norm(u, 1)
        new_y = np.sign(v[0]) * v / norm(v, 1)

        check = norm(new_x - x) + norm(new_y - y)
        if check < tol:
            break
        x = new_x.copy()
        y = new_y.copy()
    else:
        warn("Iteration did not converge!")
    return {node_dict[n]: new_x[n] for n in node_dict}, {
        edge_dict[e]: new_y[e] for e in edge_dict
    }


def line_vector_centrality(H):
    """The vector centrality of nodes in the line graph of the hypergraph.

    Parameters
    ----------
    H : Hypergraph
        The hypergraph of interest

    Returns
    -------
    dict
        Centrality, where keys are node IDs and values are lists of centralities.

    References
    ----------
    "Vector centrality in hypergraphs", K. Kovalenko, M. Romance, E. Vasilyeva,
    D. Aleja, R. Criado, D. Musatov, A.M. Raigorodskii, J. Flores, I. Samoylenko,
    K. Alfaro-Bittner, M. Perc, S. Boccaletti,
    https://doi.org/10.1016/j.chaos.2022.112397

    """
    from ..algorithms import is_connected

    # If the hypergraph is empty, then return an empty dictionary
    if H.num_nodes == 0:
        return dict()

    if not is_connected(H):
        raise XGIError("This method is not defined for disconnected hypergraphs.")

    LG = to_line_graph(H)
    LGcent = nx.eigenvector_centrality(LG)

    vc = {node: [] for node in H.nodes}

    edge_label_dict = {tuple(edge): index for index, edge in H._edge.items()}

    hyperedge_dims = {tuple(edge): len(edge) for edge in H.edges.members()}

    D = H.edges.size.max()

    for k in range(2, D + 1):
        c_i = np.zeros(len(H.nodes))

        for edge, _ in list(filter(lambda x: x[1] == k, hyperedge_dims.items())):
            for node in edge:
                try:
                    c_i[node] += LGcent[edge_label_dict[edge]]
                except IndexError:
                    raise Exception(
                        "Nodes must be written with the Pythonic indexing (0,1,2...)"
                    )

        c_i *= 1 / k

        for node in H.nodes:
            vc[node].append(c_i[node])

    return vc


def katz_centrality(H, index=False, cutoff=100):
    r"""Returns the Katz-centrality vector of a non-empty hypergraph H.

    The Katz-centrality measures the relative importance of a node by counting
    how many distinct walks start from it. The longer the walk is the smaller
    its contribution will be (attenuation factor :math:`\alpha`).
    Initially defined for graphs, the Katz-centrality is here generalized to
    hypergraphs using the most basic definition of neighbors: two nodes that
    share an hyperedge.

    Parameters
    ----------
    H : xgi.Hypergraph
        Hypergraph on which to compute the Katz-centralities.
    index : bool
        If set to `True`, will return a dictionary mapping each vector index to a
        node. Default value is `False`.
    cutoff : int
        Power at which to stop the series :math:`A + \alpha A^2 + \alpha^2 A^3 + \dots`
        Default value is 100.

    Returns
    -------
    c : np.ndarray
        Vector of the node centralities, sorted by the node indexes.
    nodedict : dict
        If index is set to True, nodedict will contain the nodes ids, keyed by
        their indice in vector `c`.
        Thus, `c[key]` will be the centrality of node `nodedict[key]`.

    Raises
    ------
    XGIError
        If the hypergraph is empty.

    Notes
    -----
    [1] The Katz-centrality is defined as

    .. math::
        c = [(I - \alpha A^{t})^{-1} - I]{\bf 1},

    where :math:`A` is the adjency matrix of the the (hyper)graph.
    Since :math:`A^{t} = A` for undirected graphs (our case), we have:


    .. math::
        &[I + A + \alpha A^2 + \alpha^2 A^3 + \dots](I - \alpha A^{t})

        & = [I + A + \alpha A^2 + \alpha^2 A^3 + \dots](I - \alpha A)

        & = (I + A + \alpha A^2 + \alpha^2 A^3 + \dots) - A - \alpha A^2

        & - \alpha^2 A^3 - \alpha^3 A^4 - \dots

        & = I

    And :math:`(I - \alpha A^{t})^{-1} = I + A + \alpha A^2 + \alpha^2 A^3 + \dots`
    Thus we can use the power serie to compute the Katz-centrality.
    [2] The Katz-centrality of isolated nodes (no hyperedges contains them) is
    zero. The Katz-centrality of an empty hypergraph is not defined.

    References
    ----------
    See https://en.wikipedia.org/wiki/Katz_centrality#Alpha_centrality (visited
    May 20 2023) for a clear definition of Katz centrality.
    """

    if index:
        A, nodedict = clique_motif_matrix(H, index=True)
    else:
        A = clique_motif_matrix(H, index=False)

    N = len(H.nodes)
    M = len(H.edges)
    if N == 0:  # no nodes
        raise XGIError("The Katz-centrality of an empty hypergraph is not defined.")
    elif M == 0:
        c = np.zeros(N)
    else:  # there is at least one edge, both N and M are non-zero
        alpha = 1 / 2**N
        mat = A
        for power in range(1, cutoff):
            mat = alpha * mat.dot(A) + A
        u = 1 / N * np.ones(N)
        c = mat.dot(u)

    if index:
        return c, nodedict
    else:
        return c



"""Node statistics.

This module is part of the stats package, and it defines node-level statistics.  That
is, each function defined in this module is assumed to define a node-quantity mapping.
Each callable defined here is accessible via a `Network` object, or a
:class:`~xgi.core.reportviews.NodeView` object.  For more details, see the `tutorial
<https://github.com/ComplexGroupInteractions/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

Examples
--------

>>> import xgi
>>> H = xgi.DiHypergraph([[{1, 2}, {5, 6}], [{4}, {1, 3}]])
>>> H.degree()
{1: 2, 2: 1, 5: 1, 6: 1, 4: 1, 3: 1}
>>> H.nodes.degree.asdict()
{1: 2, 2: 1, 5: 1, 6: 1, 4: 1, 3: 1}

"""

__all__ = [
    "attrs",
    "degree",
    "in_degree",
    "out_degree",
]


def attrs(net, bunch, attr=None, missing=None):
    """Access node attributes.

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Nodes in `net`.
    attr : str | None (default)
        If None, return all attributes.  Otherwise, return a single attribute with name
        `attr`.
    missing : Any
        Value to impute in case a node does not have an attribute with name `attr`.
        Default is None.

    Returns
    -------
    dict
        If attr is None, return a nested dict of the form `{node: {"attr": val}}`.
        Otherwise, return a simple dict of the form `{node: val}`.

    Notes
    -----
    When requesting all attributes (i.e. when `attr` is None), no value is imputed.

    Examples
    --------
    >>> import xgi
    >>> H = xgi.DiHypergraph()
    >>> H.add_nodes_from([
    ...         (1, {"color": "red", "name": "horse"}),
    ...         (2, {"color": "blue", "name": "pony"}),
    ...         (3, {"color": "yellow", "name": "zebra"}),
    ...         (4, {"color": "red", "name": "orangutan", "age": 20}),
    ...         (5, {"color": "blue", "name": "fish", "age": 2}),
    ...     ])

    Access all attributes as different types.

    >>> H.nodes.attrs.asdict() # doctest: +NORMALIZE_WHITESPACE
    {1: {'color': 'red', 'name': 'horse'},
     2: {'color': 'blue', 'name': 'pony'},
     3: {'color': 'yellow', 'name': 'zebra'},
     4: {'color': 'red', 'name': 'orangutan', 'age': 20},
     5: {'color': 'blue', 'name': 'fish', 'age': 2}}
    >>> H.nodes.attrs.asnumpy() # doctest: +NORMALIZE_WHITESPACE
    array([{'color': 'red', 'name': 'horse'},
           {'color': 'blue', 'name': 'pony'},
           {'color': 'yellow', 'name': 'zebra'},
           {'color': 'red', 'name': 'orangutan', 'age': 20},
           {'color': 'blue', 'name': 'fish', 'age': 2}],
          dtype=object)

    Access a single attribute as different types.

    >>> H.nodes.attrs('color').asdict()
    {1: 'red', 2: 'blue', 3: 'yellow', 4: 'red', 5: 'blue'}
    >>> H.nodes.attrs('color').aslist()
    ['red', 'blue', 'yellow', 'red', 'blue']

    By default, None is imputed when a node does not have the requested attribute.

    >>> H.nodes.attrs('age').asdict()
    {1: None, 2: None, 3: None, 4: 20, 5: 2}

    Use `missing` to change the imputed value.

    >>> H.nodes.attrs('age', missing=100).asdict()
    {1: 100, 2: 100, 3: 100, 4: 20, 5: 2}

    """
    if isinstance(attr, str):
        return {n: net._node_attr[n].get(attr, missing) for n in bunch}
    elif attr is None:
        return {n: net._node_attr[n] for n in bunch}
    else:
        raise ValueError('"attr" must be str or None')


def degree(net, bunch, order=None, weight=None):
    """Node degree.

    The degree of a node is the number of edges it belongs to,
    regardless of whether it is in the head or the tail.

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Nodes in `net`.
    order : int | None
        If not None (default), only count the edges of the given order.
    weight : str | None
        If not None, specifies the name of the edge attribute that determines the weight
        of each edge.

    Returns
    -------
    dict

    Examples
    --------
    >>> import xgi
    >>> H = xgi.DiHypergraph([[{1, 2}, {5, 6}], [{4}, {1, 3}]])
    >>> H.nodes.degree.asdict()
    {1: 2, 2: 1, 5: 1, 6: 1, 4: 1, 3: 1}
    """
    if order is None and weight is None:
        return {n: len(net._node_in[n].union(net._node_out[n])) for n in bunch}
    if order is None and weight:
        return {
            n: sum(
                net._edge_attr[e].get(weight, 1)
                for e in net._node_in[n].union(net._node_out[n])
            )
            for n in bunch
        }
    if order is not None and weight is None:
        return {
            n: sum(
                1
                for e in net._node_in[n].union(net._node_out[n])
                if len(net._edge_in[e].union(net._edge_out[e])) == order + 1
            )
            for n in bunch
        }
    if order is not None and weight:
        return {
            n: sum(
                net._edge_attr[e].get(weight, 1)
                for e in net._node_in[n].union(net._node_out[n])
                if len(net._edge_in[e].union(net._edge_out[e])) == order + 1
            )
            for n in bunch
        }


def in_degree(net, bunch, order=None, weight=None):
    """Node in-degree.

    The in-degree of a node is the number of edges for which
    the node is in the head (it is a receiver).

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Nodes in `net`.
    order : int | None
        If not None (default), only count the edges of the given order.
    weight : str | None
        If not None, specifies the name of the edge attribute that determines the weight
        of each edge.

    Returns
    -------
    dict

    Examples
    --------
    >>> import xgi
    >>> H = xgi.DiHypergraph([[{1, 2}, {5, 6}], [{4}, {1, 3}]])
    >>> H.nodes.in_degree.asdict()
    {1: 1, 2: 1, 5: 0, 6: 0, 4: 1, 3: 0}
    """
    if order is None and weight is None:
        return {n: len(net._node_in[n]) for n in bunch}
    if order is None and weight:
        return {
            n: sum(net._edge_attr[e].get(weight, 1) for e in net._node_in[n])
            for n in bunch
        }
    if order is not None and weight is None:
        return {
            n: sum(
                1
                for e in net._node_in[n]
                if len(net._edge_in[e].union(net._edge_out[e])) == order + 1
            )
            for n in bunch
        }
    if order is not None and weight:
        return {
            n: sum(
                net._edge_attr[e].get(weight, 1)
                for e in net._node_in[n]
                if len(net._edge_in[e].union(net._edge_out[e])) == order + 1
            )
            for n in bunch
        }


def out_degree(net, bunch, order=None, weight=None):
    """Node out-degree.

    The out-degree of a node is the number of edges for which
    the node is in the tail (it is a sender).

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Nodes in `net`.
    order : int | None
        If not None (default), only count the edges of the given order.
    weight : str | None
        If not None, specifies the name of the edge attribute that determines the weight
        of each edge.

    Returns
    -------
    dict

    Examples
    --------
    >>> import xgi
    >>> H = xgi.DiHypergraph([[{1, 2}, {5, 6}], [{4}, {1, 3}]])
    >>> H.nodes.out_degree.asdict()
    {1: 1, 2: 0, 5: 1, 6: 1, 4: 0, 3: 1}
    """
    if order is None and weight is None:
        return {n: len(net._node_out[n]) for n in bunch}
    if order is None and weight:
        return {
            n: sum(net._edge_attr[e].get(weight, 1) for e in net._node_out[n])
            for n in bunch
        }
    if order is not None and weight is None:
        return {
            n: sum(
                1
                for e in net._node_out[n]
                if len(net._edge_in[e].union(net._edge_out[e])) == order + 1
            )
            for n in bunch
        }
    if order is not None and weight:
        return {
            n: sum(
                net._edge_attr[e].get(weight, 1)
                for e in net._node[n]
                if len(net._edge_in[e].union(net._edge_out[e])) == order + 1
            )
            for n in bunch
        }



"""Node statistics.

This module is part of the stats package, and it defines node-level statistics.  That
is, each function defined in this module is assumed to define a node-quantity mapping.
Each callable defined here is accessible via a `Network` object, or a
:class:`~xgi.core.reportviews.NodeView` object.  For more details, see the `tutorial
<https://github.com/ComplexGroupInteractions/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

Examples
--------

>>> import xgi
>>> H = xgi.Hypergraph([[1, 2, 3], [2, 3, 4, 5], [3, 4, 5]])
>>> H.degree()
{1: 1, 2: 2, 3: 3, 4: 2, 5: 2}
>>> H.nodes.degree.asdict()
{1: 1, 2: 2, 3: 3, 4: 2, 5: 2}

"""

import numpy as np

import xgi

__all__ = [
    "attrs",
    "degree",
    "average_neighbor_degree",
    "local_clustering_coefficient",
    "clustering_coefficient",
    "two_node_clustering_coefficient",
    "clique_eigenvector_centrality",
    "h_eigenvector_centrality",
    "node_edge_centrality",
]


def attrs(net, bunch, attr=None, missing=None):
    """Access node attributes.

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Nodes in `net`.
    attr : str | None (default)
        If None, return all attributes.  Otherwise, return a single attribute with name
        `attr`.
    missing : Any
        Value to impute in case a node does not have an attribute with name `attr`.
        Default is None.

    Returns
    -------
    dict
        If attr is None, return a nested dict of the form `{node: {"attr": val}}`.
        Otherwise, return a simple dict of the form `{node: val}`.

    Notes
    -----
    When requesting all attributes (i.e. when `attr` is None), no value is imputed.

    Examples
    --------
    >>> import xgi
    >>> H = xgi.Hypergraph([[1, 2, 3], [2, 3, 4, 5], [3, 4, 5]])
    >>> H.add_nodes_from([
    ...         (1, {"color": "red", "name": "horse"}),
    ...         (2, {"color": "blue", "name": "pony"}),
    ...         (3, {"color": "yellow", "name": "zebra"}),
    ...         (4, {"color": "red", "name": "orangutan", "age": 20}),
    ...         (5, {"color": "blue", "name": "fish", "age": 2}),
    ...     ])

    Access all attributes as different types.

    >>> H.nodes.attrs.asdict() # doctest: +NORMALIZE_WHITESPACE
    {1: {'color': 'red', 'name': 'horse'},
     2: {'color': 'blue', 'name': 'pony'},
     3: {'color': 'yellow', 'name': 'zebra'},
     4: {'color': 'red', 'name': 'orangutan', 'age': 20},
     5: {'color': 'blue', 'name': 'fish', 'age': 2}}
    >>> H.nodes.attrs.asnumpy() # doctest: +NORMALIZE_WHITESPACE
    array([{'color': 'red', 'name': 'horse'},
           {'color': 'blue', 'name': 'pony'},
           {'color': 'yellow', 'name': 'zebra'},
           {'color': 'red', 'name': 'orangutan', 'age': 20},
           {'color': 'blue', 'name': 'fish', 'age': 2}],
          dtype=object)

    Access a single attribute as different types.

    >>> H.nodes.attrs('color').asdict()
    {1: 'red', 2: 'blue', 3: 'yellow', 4: 'red', 5: 'blue'}
    >>> H.nodes.attrs('color').aslist()
    ['red', 'blue', 'yellow', 'red', 'blue']

    By default, None is imputed when a node does not have the requested attribute.

    >>> H.nodes.attrs('age').asdict()
    {1: None, 2: None, 3: None, 4: 20, 5: 2}

    Use `missing` to change the imputed value.

    >>> H.nodes.attrs('age', missing=100).asdict()
    {1: 100, 2: 100, 3: 100, 4: 20, 5: 2}

    """
    if isinstance(attr, str):
        return {n: net._node_attr[n].get(attr, missing) for n in bunch}
    elif attr is None:
        return {n: net._node_attr[n] for n in bunch}
    else:
        raise ValueError('"attr" must be str or None')


def degree(net, bunch, order=None, weight=None):
    """Node degree.

    The degree of a node is the number of edges it belongs to.

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Nodes in `net`.
    order : int | None
        If not None (default), only count the edges of the given order.
    weight : str | None
        If not None, specifies the name of the edge attribute that determines the weight
        of each edge.

    Returns
    -------
    dict

    """
    if order is None and weight is None:
        return {n: len(net._node[n]) for n in bunch}
    if order is None and weight:
        return {
            n: sum(net._edge_attr[e].get(weight, 1) for e in net._node[n])
            for n in bunch
        }
    if order is not None and weight is None:
        return {
            n: len([e for e in net._node[n] if len(net._edge[e]) == order + 1])
            for n in bunch
        }
    if order is not None and weight:
        return {
            n: sum(
                net._edge_attr[e].get(weight, 1)
                for e in net._node[n]
                if len(net._edge[e]) == order + 1
            )
            for n in bunch
        }


def average_neighbor_degree(net, bunch):
    """Average neighbor degree.

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Nodes in `net`.

    Returns
    -------
    dict

    Examples
    --------
    >>> import xgi, numpy as np
    >>> H = xgi.Hypergraph([[1, 2, 3], [2, 3, 4, 5], [3, 4, 5]])
    >>> np.round(H.nodes.average_neighbor_degree.asnumpy(), 3)
    array([2.5  , 2.   , 1.75 , 2.333, 2.333])

    """
    result = {}
    for n in bunch:
        neighbors = net.nodes.neighbors(n)
        result[n] = sum(len(net._node[nbr]) for nbr in neighbors)
        result[n] = result[n] / len(neighbors) if neighbors else 0
    return result


def clustering_coefficient(net, bunch):
    """Local clustering coefficient.

    This clustering coefficient is defined as the
    clustering coefficient of the unweighted pairwise
    projection of the hypergraph, i.e., `num / denom`,
    where `num` equals `A^3[n, n]` and `denom` equals
    `nu*(nu-1)/2`.  Here `A` is the adjacency matrix
    of the network and `nu` is the number of pairwise
    neighbors of `n`.

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Nodes in `net`.

    Returns
    -------
    dict

    Notes
    -----
    This is a direct generalization of the definition of local clustering coefficient
    for graphs.  It has not been tested on hypergraphs.

    Examples
    --------
    >>> import xgi, numpy as np
    >>> H = xgi.Hypergraph([[1, 2, 3], [2, 3, 4, 5], [3, 4, 5]])
    >>> H.nodes.two_node_clustering_coefficient.asnumpy()
    array([0.41666667, 0.45833333, 0.58333333, 0.66666667, 0.66666667])

    """
    cc = xgi.clustering_coefficient(net)
    return {n: cc[n] for n in cc if n in bunch}


def local_clustering_coefficient(net, bunch):
    """Compute the local clustering coefficient.

    This clustering coefficient is based on the
    overlap of the edges connected to a given node,
    normalized by the size of the node's neighborhood.

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Nodes in `net`.

    Returns
    -------
    dict
        keys are node IDs and values are the
        clustering coefficients.

    References
    ----------
    "Properties of metabolic graphs: biological organization or representation
    artifacts?"  by Wanding Zhou and Luay Nakhleh.
    https://doi.org/10.1186/1471-2105-12-132

    "Hypergraphs for predicting essential genes using multiprotein complex data"
    by Florian Klimm, Charlotte M. Deane, and Gesine Reinert.
    https://doi.org/10.1093/comnet/cnaa028

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_hypergraph(3, [1, 1])
    >>> H.nodes.local_clustering_coefficient.asdict()
    {0: 1.0, 1: 1.0, 2: 1.0}

    """
    cc = xgi.local_clustering_coefficient(net)
    return {n: cc[n] for n in cc if n in bunch}


def two_node_clustering_coefficient(net, bunch, kind="union"):
    """Return the clustering coefficients for
    each node in a Hypergraph.

    This definition averages over all of the
    two-node clustering coefficients involving the node.

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Nodes in `net`.
    kind : str
        The type of two-node clustering coefficient.
        Types are:

        * "union"
        * "min"
        * "max"

    by default, "union".

    Returns
    -------
    dict
        nodes are keys, clustering coefficients are values.

    References
    ----------
    "Clustering Coefficients in Protein Interaction Hypernetworks"
    by Suzanne Gallagher and Debra Goldberg.
    DOI: 10.1145/2506583.2506635

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_hypergraph(3, [1, 1])
    >>> H.nodes.two_node_clustering_coefficient.asdict()
    {0: 0.5, 1: 0.5, 2: 0.5}
    """
    cc = xgi.two_node_clustering_coefficient(net, kind=kind)
    return {n: cc[n] for n in cc if n in bunch}


def clique_eigenvector_centrality(net, bunch, tol=1e-6):
    """Compute the clique motif eigenvector centrality of a hypergraph.

    Parameters
    ----------
    net : xgi.Hypergraph
        The hypergraph of interest.
    bunch : Iterable
        Nodes in `net`.
    tol : float > 0, default: 1e-6
        The desired L2 error in the centrality vector.

    Returns
    -------
    dict
        Centrality, where keys are node IDs and values are centralities.

    References
    ----------
    Three Hypergraph Eigenvector Centralities,
    Austin R. Benson,
    https://doi.org/10.1137/18M1203031
    """
    c = xgi.clique_eigenvector_centrality(net, tol)
    return {n: c[n] for n in c if n in bunch}


def h_eigenvector_centrality(net, bunch, max_iter=10, tol=1e-6):
    """Compute the H-eigenvector centrality of a hypergraph.

    Parameters
    ----------
    net : xgi.Hypergraph
        The hypergraph of interest.
    bunch : Iterable
        Nodes in `net`.
    max_iter : int, default: 10
        The maximum number of iterations before the algorithm terminates.
    tol : float > 0, default: 1e-6
        The desired L2 error in the centrality vector.

    Returns
    -------
    dict
        Centrality, where keys are node IDs and values are centralities.

    References
    ----------
    Three Hypergraph Eigenvector Centralities,
    Austin R. Benson,
    https://doi.org/10.1137/18M1203031
    """
    c = xgi.h_eigenvector_centrality(net, max_iter, tol)
    return {n: c[n] for n in c if n in bunch}


def node_edge_centrality(
    net,
    bunch,
    f=lambda x: np.power(x, 2),
    g=lambda x: np.power(x, 0.5),
    phi=lambda x: np.power(x, 2),
    psi=lambda x: np.power(x, 0.5),
    max_iter=100,
    tol=1e-6,
):
    """Computes node centralities.

    Parameters
    ----------
    net : Hypergraph
        The hypergraph of interest
    bunch : Iterable
        Edges in `net`
    f : lambda function, default: x^2
        The function f as described in Tudisco and Higham.
        Must accept a numpy array.
    g : lambda function, default: x^0.5
        The function g as described in Tudisco and Higham.
        Must accept a numpy array.
    phi : lambda function, default: x^2
        The function phi as described in Tudisco and Higham.
        Must accept a numpy array.
    psi : lambda function, default: x^0.5
        The function psi as described in Tudisco and Higham.
        Must accept a numpy array.
    max_iter : int, default: 100
        Number of iterations at which the algorithm terminates
        if convergence is not reached.
    tol : float > 0, default: 1e-6
        The total allowable error in the node and edge centralities.

    Returns
    -------
    dict, dict
        The node centrality where keys are node IDs and values are associated
        centralities and the edge centrality where keys are the edge IDs and
        values are associated centralities.

    Notes
    -----
    In the paper from which this was taken, it includes general functions
    for both nodes and edges, nodes and edges may be weighted, and one can
    choose different norms for normalization, all of which are not yet
    implemented.

    This method does not output the node centralities even though they are computed.

    References
    ----------
    Node and edge nonlinear eigenvector centrality for hypergraphs,
    Francesco Tudisco & Desmond J. Higham,
    https://doi.org/10.1038/s42005-021-00704-2
    """
    c, _ = xgi.node_edge_centrality(net, f, g, phi, psi, max_iter, tol)
    return {n: c[n] for n in c if n in bunch}



"""Methods for converting to higher-order network objects."""

from copy import deepcopy

import pandas as pd
from numpy import matrix, ndarray
from scipy.sparse import (
    coo_array,
    coo_matrix,
    csc_array,
    csc_matrix,
    csr_array,
    csr_matrix,
    lil_array,
    lil_matrix,
)

from ..core import DiHypergraph, Hypergraph, SimplicialComplex
from ..exception import XGIError
from ..generators import empty_dihypergraph, empty_hypergraph, empty_simplicial_complex
from .hyperedges import from_hyperedge_dict, from_hyperedge_list
from .incidence import from_incidence_matrix
from .pandas import from_bipartite_pandas_dataframe
from .simplex import from_simplex_dict

__all__ = [
    "convert_to_hypergraph",
    "convert_to_dihypergraph",
    "convert_to_simplicial_complex",
]


def convert_to_hypergraph(data, create_using=None):
    """Make a hypergraph from a known data structure.

    The preferred way to call this is automatically from the class constructor.

    Parameters
    ----------
    data : object to be converted
        Current known types are:
         * a Hypergraph object
         * a SimplicialComplex object
         * list-of-iterables
         * dict-of-iterables
         * Pandas DataFrame (bipartite edgelist)
         * numpy matrix
         * numpy ndarray
         * scipy sparse matrix
    create_using : Hypergraph constructor, optional (default=Hypergraph)
        Hypergraph type to create. If hypergraph instance, then cleared before
        populated.

    Returns
    -------
    Hypergraph object
        A hypergraph constructed from the data

    See Also
    --------
    from_max_simplices : Constructs a hypergraph from the maximal simplices of a simplicial complex.

    """
    if data is None:
        return empty_hypergraph(create_using)

    elif isinstance(data, Hypergraph) and not isinstance(data, SimplicialComplex):
        H = empty_hypergraph(create_using)
        H.add_nodes_from((n, attr) for n, attr in data.nodes.items())
        ee = data.edges
        H.add_edges_from((ee.members(e), e, deepcopy(attr)) for e, attr in ee.items())
        H._hypergraph = deepcopy(data._hypergraph)
        return H

    elif isinstance(data, DiHypergraph):
        H = empty_hypergraph(create_using)
        H.add_nodes_from((n, attr) for n, attr in data.nodes.items())
        ee = data.edges
        H.add_edges_from((ee.members(e), e, deepcopy(attr)) for e, attr in ee.items())
        H._hypergraph = deepcopy(data._hypergraph)
        if not isinstance(create_using, DiHypergraph):
            return H

    elif isinstance(data, SimplicialComplex):
        H = empty_hypergraph(create_using)
        H.add_nodes_from((n, attr) for n, attr in data.nodes.items())
        ee = data.edges
        H.add_edges_from((ee.members(e), e, deepcopy(attr)) for e, attr in ee.items())
        H._hypergraph = deepcopy(data._hypergraph)
        return H

    elif isinstance(data, list):
        # edge list
        result = from_hyperedge_list(data, create_using)
        if not isinstance(create_using, Hypergraph):
            return result

    elif isinstance(data, pd.DataFrame):
        result = from_bipartite_pandas_dataframe(data, create_using)
        if not isinstance(create_using, Hypergraph):
            return result

    elif isinstance(data, dict):
        # edge dict in the form we need
        result = from_hyperedge_dict(data, create_using)
        if not isinstance(create_using, Hypergraph):
            return result

    elif isinstance(
        data,
        (
            ndarray,
            matrix,
            csr_array,
            csc_array,
            coo_array,
            lil_array,
            csr_matrix,
            csc_matrix,
            coo_matrix,
            lil_matrix,
        ),
    ):
        from_incidence_matrix(data, create_using)

    else:
        raise XGIError("Input data has unsupported type.")


def convert_to_dihypergraph(data, create_using=None):
    """Make a dihypergraph from a known data structure.

    The preferred way to call this is automatically from the class constructor.

    Parameters
    ----------
    data : object to be converted
        Current known types are:
         * a DiHypergraph object
         * a SimplicialComplex object
         * list-of-iterables
         * dict-of-iterables
         * Pandas DataFrame (bipartite edgelist)
         * numpy matrix
         * numpy ndarray
         * scipy sparse matrix
    create_using : Hypergraph constructor, optional (default=Hypergraph)
        Hypergraph type to create. If hypergraph instance, then cleared before populated.

    Returns
    -------
    Hypergraph object
        A hypergraph constructed from the data

    """
    if data is None:
        return empty_dihypergraph(create_using)

    elif isinstance(data, DiHypergraph):
        H = empty_dihypergraph(create_using)
        H.add_nodes_from((n, attr) for n, attr in data.nodes.items())
        ee = data.edges
        H.add_edges_from((ee.dimembers(e), e, deepcopy(attr)) for e, attr in ee.items())
        H._hypergraph = deepcopy(data._hypergraph)
        if not isinstance(create_using, DiHypergraph):
            return H

    elif isinstance(data, list):
        # edge list
        result = from_hyperedge_list(data, create_using)
        if not isinstance(create_using, DiHypergraph):
            return result

    elif isinstance(data, dict):
        # edge dict in the form we need
        result = from_hyperedge_dict(data, create_using)
        if not isinstance(create_using, DiHypergraph):
            return result

    else:
        raise XGIError("Input data has unsupported type.")


from ..generators import empty_simplicial_complex


def convert_to_simplicial_complex(data, create_using=None):
    """Make a hypergraph from a known data structure.
    The preferred way to call this is automatically
    from the class constructor.

    Parameters
    ----------
    data : object to be converted
        Current known types are:
         * a SimplicialComplex object
         * a Hypergraph object
         * list-of-iterables
         * dict-of-iterables
         * Pandas DataFrame (bipartite edgelist)
         * numpy matrix
         * numpy ndarray
         * scipy sparse matrix
    create_using : Hypergraph graph constructor, optional (default=Hypergraph)
        Hypergraph type to create. If hypergraph instance, then cleared before
        populated.

    Returns
    -------
    Hypergraph object
        A hypergraph constructed from the data

    """

    if data is None:
        return empty_simplicial_complex(create_using)

    elif isinstance(data, SimplicialComplex):
        H = empty_simplicial_complex(create_using)
        H.add_nodes_from((n, attr) for n, attr in data.nodes.items())
        ee = data.edges
        H.add_simplices_from(
            (ee.members(e), e, deepcopy(attr)) for e, attr in ee.items()
        )
        H._hypergraph = deepcopy(data._hypergraph)
        return H

    elif isinstance(data, Hypergraph):
        H = empty_simplicial_complex(create_using)
        H.add_nodes_from((n, attr) for n, attr in data.nodes.items())
        ee = data.edges
        H.add_simplices_from(
            (ee.members(e), e, deepcopy(attr)) for e, attr in ee.items()
        )
        return H

    elif isinstance(data, list):
        # edge list
        result = from_hyperedge_list(data, create_using)
        if not isinstance(create_using, SimplicialComplex):
            return convert_to_simplicial_complex(result)

    elif isinstance(data, pd.DataFrame):
        result = from_bipartite_pandas_dataframe(data, create_using)
        if not isinstance(create_using, SimplicialComplex):
            return convert_to_simplicial_complex(result)

    elif isinstance(data, dict):
        result = from_simplex_dict(data, create_using)
        if not isinstance(create_using, SimplicialComplex):
            return convert_to_simplicial_complex(result)
    elif isinstance(
        data,
        (
            ndarray,
            matrix,
            csr_array,
            csc_array,
            coo_array,
            lil_array,
            csr_matrix,
            csc_matrix,
            coo_matrix,
            lil_matrix,
        ),
    ):
        # incidence matrix
        raise XGIError(
            "Not implemented: construction of a SimplicialComplex from incidence matrix"
        )
    else:
        raise XGIError("Input data has unsupported type.")



"""General utilities."""

from collections import defaultdict
from copy import deepcopy
from functools import lru_cache
from itertools import chain, combinations, count

import numpy as np
import pandas as pd
import requests
from numpy import infty

from xgi.exception import IDNotFound, XGIError

__all__ = [
    "IDDict",
    "dual_dict",
    "powerset",
    "update_uid_counter",
    "find_triangles",
    "request_json_from_url",
    "request_json_from_url_cached",
    "subfaces",
    "convert_labels_to_integers",
    "hist",
    "binomial_sequence",
]


class IDDict(dict):
    """A dict that holds (node or edge) IDs.

    For internal use only.  Adds input validation functionality to the internal dicts
    that hold nodes and edges in a network.

    """

    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError as e:
            raise IDNotFound(f"ID {item} not found") from e

    def __setitem__(self, item, value):
        if item is None:
            raise XGIError("None cannot be a node or edge")
        try:
            return dict.__setitem__(self, item, value)
        except TypeError as e:
            raise TypeError(f"ID {item} not a valid type") from e

    def __delitem__(self, item):
        try:
            return dict.__delitem__(self, item)
        except KeyError as e:
            raise IDNotFound(f"ID {item} not found") from e


def dual_dict(edge_dict):
    """Given a dictionary with IDs as keys
    and sets as values, return the dual.

    Parameters
    ----------
    edge_dict : dict
        A dictionary where the keys are
        IDs and the values are sets of hashables

    Returns
    -------
    dict
        A dictionary with IDs as keys
        and sets as values, but the reverse of
        the original dict.

    Examples
    --------
    >>> import xgi
    >>> xgi.dual_dict({0 : [1, 2, 3], 1 : [0, 2]})
    {1: {0}, 2: {0, 1}, 3: {0}, 0: {1}}

    """
    node_dict = defaultdict(set)
    for edge_id, members in edge_dict.items():
        for node in members:
            node_dict[node].add(edge_id)

    return dict(node_dict)


def powerset(
    iterable,
    include_empty=False,
    include_full=False,
    include_singletons=True,
    max_size=None,
):
    """Returns all possible subsets of the elements in iterable, with options
    to include the empty set and the set containing all elements, and to set
    the maximum subset size.

    Parameters
    ----------
    iterable : list-like
        List of elements
    include_empty: bool, default: False
        Whether to include the empty set
    include_singletons: bool, default: True
        Whether to include singletons
    include_full: bool, default: False
        Whether to include the set containing all elements of iterable
    max_size: int, default: None
        Maximum size of the returned subsets.

    Returns
    -------
    itertools.chain

    Notes
    -----
    include_empty overrides include_singletons if True: singletons will always
    be included if the empty set is.
    Likewise, max_size will override other arguments: if set to -1, no subset
    will be returned.

    Examples
    --------
    >>> import xgi
    >>> list(xgi.powerset([1,2,3,4])) # doctest: +NORMALIZE_WHITESPACE
    [(1,), (2,), (3,), (4,), (1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4),
     (1, 2, 3), (1, 2, 4), (1, 3, 4), (2, 3, 4)]

    """

    start = 1 if include_singletons else 2
    start = 0 if include_empty else start  # overrides include_singletons if True

    s = list(iterable)

    if max_size is None:
        max_size = len(s) if include_full else len(s) - 1
    else:
        max_size = min(max_size, len(s))

    return chain.from_iterable(combinations(s, r) for r in range(start, max_size + 1))


def update_uid_counter(H, new_id):
    """
    Helper function to make sure the uid counter is set correctly after
    adding an edge with a user-provided ID.

    If we don't set the start of self._edge_uid correctly, it will start at 0,
    which will overwrite any existing edges when calling add_edge().  First, we
    use the somewhat convoluted float(e).is_integer() instead of using
    isinstance(e, int) because there exist integer-like numeric types (such as
    np.int32) which fail the isinstance() check.

    Parameters
    ----------
    H : xgi.Hypergraph
        Hypergraph of which to update the uid counter
    id : any hashable type
        User-provided ID.

    """
    uid = next(H._edge_uid)
    if (
        not isinstance(new_id, str)
        and not isinstance(new_id, tuple)
        and float(new_id).is_integer()
        and uid <= new_id
    ):
        # tuple comes from merging edges and doesn't have as as_integer() method.
        start = int(new_id) + 1
        # we set the start at one plus the maximum edge ID that is an integer,
        # because count() only yields integer IDs.
    else:
        start = uid
    H._edge_uid = count(start=start)


def find_triangles(G):
    """Returns list of 3-node cliques present in a graph

    Parameters
    ----------
    G : networkx Graph
        Graph to consider

    Returns
    -------
    list of 3-node cliques (triangles)
    """

    triangles = set(
        frozenset((n, nbr, nbr2))
        for n in G
        for nbr, nbr2 in combinations(G[n], 2)
        if nbr in G[nbr2]
    )
    return [set(tri) for tri in triangles]


def min_where(dicty, where):
    """
    Finds the minimum value of a dictonary `dicty`. The dictonary `where`
    indicates which keys to take into account. The minimum is eventualy
    infinite.

    Parameters
    ----------
    dicty : dict
        Dictionary of values (int, float...) from which to find
        the minimum.
    where : dict
        Dictionary of booleans that has the same keys as `dicty`.
        The minimum will be searched among the values for which
        `where[key]` is TRUE.

    Return
    ------
    min_val : float or np.Inf
        Minimum value found in `dicty`. Is set to np.infty if `where` indicated
        nowhere or if all values are `np.infty`.
    """

    min_val = infty
    for key in dicty.keys():
        if where[key]:
            if dicty[key] < min_val:
                min_val = dicty[key]
            else:
                pass
        else:
            pass
    return min_val


def request_json_from_url(url):
    """HTTP request json file and return as dict.

    Parameters
    ----------
    url : str
        The url where the json file is located.

    Returns
    -------
    dict
        A dictionary of the JSON requested.

    Raises
    ------
    XGIError
        If the connection fails or if there is a bad HTTP request.
    """

    try:
        r = requests.get(url)
    except requests.ConnectionError:
        raise XGIError("Connection Error!")

    if r.ok:
        return r.json()
    else:
        raise XGIError(f"Error: HTTP response {r.status_code}")


@lru_cache(maxsize=None)
def request_json_from_url_cached(url):
    """HTTP request json file and return as dict.

    Parameters
    ----------
    url : str
        The url where the json file is located.

    Returns
    -------
    dict
        A dictionary of the JSON requested.

    Raises
    ------
    XGIError
        If the connection fails or if there is a bad HTTP request.
    """

    try:
        r = requests.get(url)
    except requests.ConnectionError:
        raise XGIError("Connection Error!")

    if r.ok:
        return r.json()
    else:
        raise XGIError(f"Error: HTTP response {r.status_code}")


def subfaces(edges, order=None):
    """Returns the subfaces of a list of hyperedges

    Parameters
    ---------
    edges: list of edges
        Edges to consider, as tuples of nodes
    order: {None, -1, int}, optional
        If None, compute subfaces recursively down to nodes.
        If -1, compute subfaces the order below (e.g. edges for a triangle).
        If d > 0, compute the subfaces of order d.
        By default, None.

    Returns
    -------
    faces: list of sets
        List of hyperedges that are subfaces of input hyperedges.

    Raises
    ------
    XGIError
        Raises error when order is larger than the max order of input edges

    Notes
    -----
    Hyperedges in the returned list are not unique, they may appear more than once
    if they are subfaces or more than one edge from the input edges.

    Examples
    --------
    >>> import xgi
    >>> edges = [{1,2,3,4}, {3,4,5}]
    >>> xgi.subfaces(edges) # doctest: +NORMALIZE_WHITESPACE
    [(1,), (2,), (3,), (4,), (1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4), (1, 2, 3),
     (1, 2, 4), (1, 3, 4), (2, 3, 4), (3,), (4,), (5,), (3, 4), (3, 5), (4, 5)]
    >>> xgi.subfaces(edges, order=-1)
    [(1, 2, 3), (1, 2, 4), (1, 3, 4), (2, 3, 4), (3, 4), (3, 5), (4, 5)]
    >>> xgi.subfaces(edges, order=2)
    [(1, 2, 3), (1, 2, 4), (1, 3, 4), (2, 3, 4), (3, 4, 5)]

    """

    max_order = len(max(edges, key=len)) - 1
    if order and order > max_order:
        raise XGIError(
            "order must be less or equal to the maximum "
            f"order among the edges: {max_order}."
        )

    faces = []
    for edge in edges:
        size = len(edge)

        if size <= 1:  # down from a node is an empty tuple
            continue

        if order is None:  # add all subfaces down to nodes
            faces_to_add = list(powerset(edge))
        elif order == -1:  # add subfaces of order below
            faces_to_add = list(combinations(edge, size - 1))
        elif order >= 0:  # add subfaces of order d
            faces_to_add = list(combinations(edge, order + 1))

        faces += faces_to_add
    return faces


def convert_labels_to_integers(net, label_attribute="label"):
    """Relabel node and edge IDs to be sequential integers.

    Parameters
    ----------
    net : Hypergraph, DiHypergraph, or SimplicialComplex
        The higher-order network of interest

    label_attribute : string, default: "label"
        The attribute name that stores the old node and edge labels

    Returns
    -------
    Hypergraph, DiHypergraph, or SimplicialComplex
        A new higher-order network with nodes and edges with sequential IDs starting at 0.
        The old IDs are stored in the "label" attribute for both nodes and edges.

    Notes
    -----
    The "relabeling" will occur even if the node/edge IDs are sequential.
    Because the old IDs are stored in the "label" attribute for both nodes and edges,
    the old "label" values (if they exist) will be overwritten.
    """
    from ..core import DiHypergraph, Hypergraph, SimplicialComplex

    node_dict = dict(zip(net.nodes, range(net.num_nodes)))
    edge_dict = dict(zip(net.edges, range(net.num_edges)))

    temp_net = net.__class__()
    temp_net._hypergraph = deepcopy(net._hypergraph)

    temp_net.add_nodes_from((id, deepcopy(net.nodes[n])) for n, id in node_dict.items())
    temp_net.set_node_attributes(
        {n: {label_attribute: id} for id, n in node_dict.items()}
    )
    if isinstance(net, SimplicialComplex):
        temp_net.add_simplices_from(
            (
                {node_dict[n] for n in e},
                edge_dict[id],
                deepcopy(net.edges[id]),
            )
            for id, e in net.edges.members(dtype=dict).items()
        )
    elif isinstance(net, Hypergraph):
        temp_net.add_edges_from(
            (
                {node_dict[n] for n in e},
                edge_dict[id],
                deepcopy(net.edges[id]),
            )
            for id, e in net.edges.members(dtype=dict).items()
        )
    elif isinstance(net, DiHypergraph):
        temp_net.add_edges_from(
            (
                [{node_dict[n] for n in tail}, {node_dict[n] for n in head}],
                edge_dict[id],
                deepcopy(net.edges[id]),
            )
            for id, (tail, head) in net.edges.dimembers(dtype=dict).items()
        )

    temp_net.set_edge_attributes(
        {e: {label_attribute: id} for id, e in edge_dict.items()}
    )

    return temp_net


def hist(vals, bins=10, bin_edges=False, density=False, log_binning=False):
    """Return the distribution of a numpy array.

    Parameters
    ----------
    vals : Numpy array
        The array of values
    bins : int, list, or Numpy array
        The number of bins or the bin edges.
        By default, 10.
    bin_edges : bool
        Whether to also output the min and max of each bin,
        by default, False.
    density : bool
        Whether to normalize the resulting distribution.
        By default, False.
    log_binning : bool
        Whether to bin the values with log-sized bins.
        By default, False.

    Returns
    -------
    Pandas DataFrame
        A two-column table with "bin_center" and "value" columns,
        where "value" is a count or a probability. If `bin_edges`
        is True, outputs two additional columns, `bin_lo` and `bin_hi`,
        which outputs the left and right bin edges respectively.

    Notes
    -----
    Originally from https://github.com/jkbren/networks-and-dataviz

    """
    # We need to define the support of our distribution
    lower_bound = np.min(vals)
    upper_bound = np.max(vals)

    if log_binning:
        lower_bound = np.log10(lower_bound) if lower_bound > 0 else 0.0
        upper_bound = np.log10(upper_bound)

    # And the bins
    if isinstance(bins, int):
        if log_binning:
            bins = np.logspace(lower_bound, upper_bound, bins + 1, base=10)
        else:
            bins = np.linspace(lower_bound, upper_bound, bins + 1)
    elif not isinstance(bins, (list, np.ndarray)):
        raise XGIError("Bins must be an integer, a list, or a numpy array.")

    # Then we can compute the histogram using numpy
    y, __ = np.histogram(vals, bins=bins, density=density)
    # Now, we need to compute for each y the bin centers
    x = bins[1:] - np.diff(bins) / 2.0

    if bin_edges:
        return pd.DataFrame.from_dict(
            {"bin_center": x, "value": y, "bin_lo": bins[:-1], "bin_hi": bins[1:]}
        )
    else:
        return pd.DataFrame.from_dict({"bin_center": x, "value": y})


def binomial_sequence(k, N):
    """Returns the set of all the distinct strings (order counts) with k ones
    and N-k zeros. binomial_sequence(2, 4) will output '1010', '1100', '0011', '0110',
    '0101' and '1001'.

    Parameters
    ----------
    k : int
        Number of ones in the strings. Must be greater or equal to zero.
    N : int
        Length of the strings. Must be positive as well.

    Returns
    -------
    res : set
        Set containing all the strings (they are distinct). The empty set is
        returned if N = 0 or if k > N.
    """

    if k < 0 or N < 0:
        raise ValueError("binomial_sequence must be given positive integers.")

    res = set()

    if k == 0:
        # seq = str()
        # for i in range(N):
        #    seq += "0"
        # res.add(seq)
        res = {"0" * N}
    elif N == 0:
        pass
    else:  # k and N are greater than zero
        for seq in binomial_sequence(k, N - 1):
            res.add(seq + "0")
        for seq in binomial_sequence(k - 1, N - 1):
            res.add(seq + "1")
    return res



"""Methods for converting to and from hyperedge lists."""

from ..core import SimplicialComplex
from ..generators import empty_hypergraph

__all__ = [
    "from_hyperedge_dict",
    "to_hyperedge_dict",
    "from_hyperedge_list",
    "to_hyperedge_list",
]


def from_hyperedge_dict(d, create_using=None):
    """Creates a hypergraph from a dictionary of hyperedges

    Parameters
    ----------
    d : dict
        A dictionary where the keys are edge IDs and the values
        are containers of nodes specifying the edges.
    create_using : Hypergraph constructor, optional
        The hypergraph object to add the data to, by default None

    Returns
    -------
    Hypergraph object
        The constructed hypergraph object

    See Also
    --------
    to_hyperedge_dict

    """
    H = empty_hypergraph(create_using)
    H.add_edges_from((members, uid) for uid, members in d.items())
    return H


def to_hyperedge_dict(H):
    """Outputs a hyperedge dictionary

    Parameters
    ----------
    H : Hypergraph object
        The hypergraph of interest

    Returns
    -------
    dict
        A dictionary where the keys are edge IDs and the values
        are sets of nodes specifying the edges.

    See Also
    --------
    from_hyperedge_dict
    """
    return H.edges.members(dtype=dict)


def from_hyperedge_list(d, create_using=None, max_order=None):
    """Generate a hypergraph from a list of lists.

    Parameters
    ----------
    d : list of iterables
        A hyperedge list
    create_using : Hypergraph constructor, optional
        The hypergraph to add the edges to, by default None

    Returns
    -------
    Hypergraph object
        The constructed hypergraph object

    See Also
    --------
    to_hyperedge_list
    """
    H = empty_hypergraph(create_using)
    if isinstance(H, SimplicialComplex):
        H.add_simplices_from(d, max_order=max_order)
    else:
        H.add_edges_from(d)
    return H


def to_hyperedge_list(H):
    """Generate a hyperedge list from a hypergraph.

    Parameters
    ----------
    H : Hypergraph object
        The hypergraph of interest

    Returns
    -------
    list of sets
        The hyperedge list

    See Also
    --------
    from_hyperedge_list
    """
    return H.edges.members()



"""View of Hypergraphs as a subhypergraph or read-only.

In some algorithms it is convenient to temporarily morph a hypergraph to exclude some
nodes or edges. It should be better to do that via a view than to remove and then
re-add. This module provides those graph views.

The resulting views are essentially read-only graphs that report data from the original
graph object.

Note: Since globalviews look like hypergraphs, one can end up with
view-of-view-of-view chains. Be careful with chains because they become very slow with
about 15 nested views. Often it is easiest to use .copy() to avoid chains.

"""

__all__ = ["subhypergraph"]


def subhypergraph(H, nodes=None, edges=None, keep_isolates=True):
    """View of `H` applying a filter on nodes and edges.

    `subhypergraph_view` provides a read-only view of the induced subhypergraph that
    includes nodes, edges, or both based on what the user specifies. This function
    automatically filters out edges that are not subsets of the nodes. This function may
    create isolated nodes.

    If the user only specifies the nodes to include, the function returns an induced
    subhypergraph on the nodes.

    If the user only specifies the edges to include, the function returns all of the
    nodes and the specified edges.

    If the user specifies both nodes and edges to include in the subhypergraph, then the
    function returns a subhypergraph with the specified nodes and edges from the list of
    specified hyperedges that are induced by the specified nodes.

    Parameters
    ----------
    H : hypergraph.Hypergraph
        A hypergraph
    nodes : list or set, optional
        A list of the nodes desired for the subhypergraph.
        If None (default), uses all the nodes.
    edges : list or set, optional
        A list of the edges desired for the subhypergraph.
        If None (default), uses all the edges.
    keep_isolates : bool, optional
        Whether to keep isolated nodes in the subhypergraph.
        By default, True.

    Returns
    -------
    Hypergraph object
        A read-only hypergraph view of the input hypergraph.

    """
    new = H.__class__()

    new._hypergraph = H._hypergraph.copy()
    nodes = set(H.nodes) if nodes is None else (set(nodes) & set(H.nodes))
    edges = set(H.edges) if edges is None else (set(edges) & set(H.edges))

    new.add_nodes_from((uid, attr) for uid, attr in H.nodes.items() if uid in nodes)
    new.add_edges_from(
        (H.edges.members(uid), uid, attr)
        for uid, attr in H.edges.items()
        if uid in edges and set(H.edges.members(uid)).issubset(nodes)
    )

    if not keep_isolates:
        new.remove_nodes_from(new.nodes.isolates())

    new.freeze()
    return new



"""Directed edge statistics.

This module is part of the stats package, and it defines edge-level statistics.  That
is, each function defined in this module is assumed to define an edge-quantity mapping.
Each callable defined here is accessible via a `Network` object, or a
:class:`~xgi.core.reportviews.DiEdgeView` object.  For more details, see the `tutorial
<https://github.com/xgi-org/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

Examples
--------

>>> import xgi
>>> H = xgi.DiHypergraph([[{1, 2}, {5, 6}], [{4}, {1, 3}]])
>>> H.order()
{0: 3, 1: 2}
>>> H.edges.order.asdict()
{0: 3, 1: 2}

"""

__all__ = [
    "attrs",
    "order",
    "size",
    "head_order",
    "head_size",
    "tail_order",
    "tail_size",
]


def attrs(net, bunch, attr=None, missing=None):
    """Access edge attributes.

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Nodes in `net`.
    attr : str | None (default)
        If None, return all attributes.  Otherwise, return a single attribute with name
        `attr`.
    missing : Any
        Value to impute in case an edge does not have an attribute with name `attr`.
        Default is None.

    Returns
    -------
    dict
        If attr is None, return a nested dict of the form `{edge: {"attr": val}}`.
        Otherwise, return a simple dict of the form `{edge: val}`.

    Notes
    -----
    When requesting all attributes (i.e. when `attr` is None), no value is imputed.

    Examples
    --------
    >>> import xgi
    >>> H = xgi.DiHypergraph()
    >>> edges = [
    ...     ([{0, 1}, {2, 4}], 'one', {'color': 'red'}),
    ...     ([{1, 2}, {2, 0}], 'two', {'color': 'black', 'age': 30}),
    ...     ([{2, 3, 4}, {1}], 'three', {'color': 'blue', 'age': 40}),
    ... ]
    >>> H.add_edges_from(edges)

    Access all attributes as different types.

    >>> H.edges.attrs.asdict() # doctest: +NORMALIZE_WHITESPACE
    {'one': {'color': 'red'},
     'two': {'color': 'black', 'age': 30},
     'three': {'color': 'blue', 'age': 40}}
    >>> H.edges.attrs.asnumpy() # doctest: +NORMALIZE_WHITESPACE
    array([{'color': 'red'},
           {'color': 'black', 'age': 30},
           {'color': 'blue', 'age': 40}],
           dtype=object)

    Access a single attribute as different types.

    >>> H.edges.attrs('color').asdict()
    {'one': 'red', 'two': 'black', 'three': 'blue'}
    >>> H.edges.attrs('color').aslist()
    ['red', 'black', 'blue']

    By default, None is imputed when a node does not have the requested attribute.

    >>> H.edges.attrs('age').asdict()
    {'one': None, 'two': 30, 'three': 40}

    Use `missing` to change the imputed value.

    >>> H.edges.attrs('age', missing=100).asdict()
    {'one': 100, 'two': 30, 'three': 40}

    """
    if isinstance(attr, str):
        return {e: net._edge_attr[e].get(attr, missing) for e in bunch}
    elif attr is None:
        return {e: net._edge_attr[e] for e in bunch}
    else:
        raise ValueError('"attr" must be str or None')


def order(net, bunch, degree=None):
    """Edge order.

    The order of a directed edge is the number of nodes
    contained in the union of the head and the tail minus 1.

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Edges in `net`.
    degree : int | None
        If not None (default), count only those member nodes with the specified degree.

    Returns
    -------
    dict

    See Also
    --------
    size

    Examples
    --------
    >>> import xgi
    >>> H = xgi.DiHypergraph([[{1, 2}, {5, 6}], [{4}, {1, 3}]])
    >>> H.edges.order.asdict()
    {0: 3, 1: 2}
    """
    if degree is None:
        return {e: len(net._edge_in[e].union(net._edge_out[e])) - 1 for e in bunch}
    else:
        return {
            e: sum(
                len(net._node_in[n].union(net._node_out[n])) == degree
                for n in net._edge_in[e].union(net._edge_out[e])
            )
            - 1
            for e in bunch
        }


def size(net, bunch, degree=None):
    """Edge size.

    The size of a directed edge is the number of nodes
    contained in the union of the head and the tail.

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Edges in `net`.

    Returns
    -------
    dict

    See Also
    --------
    order

    Examples
    --------
    >>> import xgi
    >>> H = xgi.DiHypergraph([[{1, 2}, {5, 6}], [{4}, {1, 3}]])
    >>> H.edges.size.asdict()
    {0: 4, 1: 3}
    """
    if degree is None:
        return {e: len(net._edge_in[e].union(net._edge_out[e])) for e in bunch}
    else:
        return {
            e: sum(
                1
                for n in net._edge_in[e].union(net._edge_out[e])
                if len(net._node_in[n].union(net._node_out[n])) == degree
            )
            for e in bunch
        }


def tail_order(net, bunch, degree=None):
    """Tail order.

    The order of the tail is the number of nodes it contains minus 1.

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Edges in `net`.

    Returns
    -------
    dict

    See Also
    --------
    order

    Examples
    --------
    Examples
    --------
    >>> import xgi
    >>> H = xgi.DiHypergraph([[{1, 2}, {5, 6}], [{4}, {1, 3}]])
    >>> H.edges.tail_order.asdict()
    {0: 1, 1: 0}

    """
    if degree is None:
        return {e: len(net._edge_out[e]) - 1 for e in bunch}
    else:
        return {
            e: sum(
                1
                for n in net._edge_out[e]
                if len(net._node_in[n].union(net._node_out[n])) == degree
            )
            - 1
            for e in bunch
        }


def tail_size(net, bunch, degree=None):
    """Tail size.

    The size of the tail is the number of nodes it contains.

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Edges in `net`.

    Returns
    -------
    dict

    See Also
    --------
    order

    Examples
    --------
    >>> import xgi
    >>> H = xgi.DiHypergraph([[{1, 2}, {5, 6}], [{4}, {1, 3}]])
    >>> H.edges.tail_size.asdict()
    {0: 2, 1: 1}
    """
    if degree is None:
        return {e: len(net._edge_out[e]) for e in bunch}
    else:
        return {
            e: sum(
                1
                for n in net._edge_out[e]
                if len(net._node_in[n].union(net._node_out[n])) == degree
            )
            for e in bunch
        }


def head_order(net, bunch, degree=None):
    """Head order.

    The order of the head is the number of nodes it contains minus 1.

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Edges in `net`.

    Returns
    -------
    dict

    See Also
    --------
    order

    Examples
    --------
    >>> import xgi
    >>> H = xgi.DiHypergraph([[{1, 2}, {5, 6}], [{4}, {1, 3}]])
    >>> H.edges.head_order.asdict()
    {0: 1, 1: 1}
    """
    if degree is None:
        return {e: len(net._edge_in[e]) - 1 for e in bunch}
    else:
        return {
            e: sum(
                1
                for n in net._edge_in[e]
                if len(net._node_in[n].union(net._node_out[n])) == degree
            )
            - 1
            for e in bunch
        }


def head_size(net, bunch, degree=None):
    """Head size.

    The size of the head is the number of nodes it contains.

    Parameters
    ----------
    net : xgi.Hypergraph
        The network.
    bunch : Iterable
        Edges in `net`.

    Returns
    -------
    dict

    See Also
    --------
    order

    Examples
    --------
    >>> import xgi
    >>> H = xgi.DiHypergraph([[{1, 2}, {5, 6}], [{4}, {1, 3}]])
    >>> H.edges.head_size.asdict()
    {0: 2, 1: 2}
    """
    if degree is None:
        return {e: len(net._edge_in[e]) for e in bunch}
    else:
        return {
            e: sum(
                1
                for n in net._edge_in[e]
                if len(net._node_in[n].union(net._node_out[n])) == degree
            )
            for e in bunch
        }



"""Hodge theory matrices associated to hypergraphs.

Note that the order of the rows and columns of the
matrices in this module correspond to the order in
which nodes/edges are added to the hypergraph or
simplicial complex. If the node and edge IDs are
able to be sorted, the following is an example to sort
by the node and edge IDs.

>>> import xgi
>>> import pandas as pd
>>> H = xgi.Hypergraph([[1, 2, 3, 7], [4], [5, 6, 7]])
>>> I, nodedict, edgedict = xgi.incidence_matrix(H, sparse=False, index=True)
>>> # Sorting the resulting numpy array:
>>> sortedI = I.copy()
>>> sortedI = sortedI[sorted(nodedict, key=nodedict.get), :]
>>> sortedI = sortedI[:, sorted(edgedict, key=edgedict.get)]
>>> sortedI
array([[1, 0, 0],
       [1, 0, 0],
       [1, 0, 0],
       [0, 1, 0],
       [0, 0, 1],
       [0, 0, 1],
       [1, 0, 1]])
>>> # Indexing a Pandas dataframe by the node/edge IDs
>>> df = pd.DataFrame(I, index=nodedict.values(), columns=edgedict.values())

If the nodes are already sorted, this order can be preserved by adding the nodes
to the hypergraph prior to adding edges. For example,

>>> import xgi
>>> H = xgi.Hypergraph()
>>> H.add_nodes_from(range(1, 8))
>>> H.add_edges_from([[1, 2, 3, 7], [4], [5, 6, 7]])
>>> xgi.incidence_matrix(H, sparse=False)
array([[1, 0, 0],
       [1, 0, 0],
       [1, 0, 0],
       [0, 1, 0],
       [0, 0, 1],
       [0, 0, 1],
       [1, 0, 1]])

"""
import numpy as np

__all__ = [
    "boundary_matrix",
    "hodge_laplacian",
]


def boundary_matrix(S, order=1, orientations=None, index=False):
    """Generate the boundary matrices of an oriented simplicial complex.

    The rows correspond to the (order-1)-simplices and the columns to the
    (order)-simplices.

    Parameters
    ----------
    S: simplicial complex object
        The simplicial complex of interest
    order: int, default: 1
        Specifies the order of the boundary
        matrix to compute
    orientations: dict, default: None
        Dictionary mapping non-singleton simplices
        IDs to their boolean orientation
    index: bool, default: False
        Specifies whether to output dictionaries
        mapping the simplices IDs to indices

    Returns
    -------
    B: numpy.ndarray
        The boundary matrix of the chosen order, has dimension
        (n_simplices of given order - 1, n_simplices of given order)
    rowdict: dict
        The dictionary mapping indices to
        (order-1)-simplices IDs, if index is True
    coldict: dict
        The dictionary mapping indices to
        (order)-simplices IDs, if index is True

    References
    ----------
    "Discrete Calculus"
    by Leo J. Grady and Jonathan R. Polimeni
    https://doi.org/10.1007/978-1-84996-290-2

    """

    # Extract the simplices involved
    if order == 1:
        simplices_d_ids = S.nodes
    else:
        simplices_d_ids = S.edges.filterby("order", order - 1)

    if order == 0:
        simplices_u_ids = S.nodes
    else:
        simplices_u_ids = S.edges.filterby("order", order)
    nd = len(simplices_d_ids)
    nu = len(simplices_u_ids)

    simplices_d_dict = dict(zip(simplices_d_ids, range(nd)))
    simplices_u_dict = dict(zip(simplices_u_ids, range(nu)))

    if index:
        rowdict = {v: k for k, v in simplices_d_dict.items()}
        coldict = {v: k for k, v in simplices_u_dict.items()}

    if orientations is None:
        orientations = {idd: 0 for idd in S.edges.filterby("order", 1, mode="geq")}

    B = np.zeros((nd, nu))
    if not (nu == 0 or nd == 0):
        if order == 1:
            for u_simplex_id in simplices_u_ids:
                u_simplex = list(S.edges.members(u_simplex_id))
                u_simplex.sort(
                    key=lambda e: (isinstance(e, str), e)
                )  # Sort the simplex's vertices to get a reference orientation
                # The key is needed to sort a mixed list of numbers and strings:
                #   it ensures that node labels which are numbers are put before
                #   strings, thus giving a list [sorted numbers, sorted strings]
                matrix_id = simplices_u_dict[u_simplex_id]
                head_idx = u_simplex[1]
                tail_idx = u_simplex[0]
                B[simplices_d_dict[head_idx], matrix_id] = (-1) ** orientations[
                    u_simplex_id
                ]
                B[simplices_d_dict[tail_idx], matrix_id] = -(
                    (-1) ** orientations[u_simplex_id]
                )
        else:
            for u_simplex_id in simplices_u_ids:
                u_simplex = list(S.edges.members(u_simplex_id))
                u_simplex.sort(
                    key=lambda e: (isinstance(e, str), e)
                )  # Sort the simplex's vertices to get a reference orientation
                # The key is needed to sort a mixed list of numbers and strings:
                #   it ensures that node labels which are numbers are put before
                #   strings, thus giving a list [sorted numbers, sorted strings]
                matrix_id = simplices_u_dict[u_simplex_id]
                u_simplex_subfaces = S._subfaces(u_simplex, all=False)
                subfaces_induced_orientation = [
                    (orientations[u_simplex_id] + order - i) % 2
                    for i in range(order + 1)
                ]
                for count, subf in enumerate(u_simplex_subfaces):
                    subface_ID = list(S.edges)[S.edges.members().index(frozenset(subf))]
                    B[simplices_d_dict[subface_ID], matrix_id] = (-1) ** (
                        subfaces_induced_orientation[count] + orientations[subface_ID]
                    )
    return (B, rowdict, coldict) if index else B


def hodge_laplacian(S, order=1, orientations=None, index=False):
    """
    A function to compute the Hodge Laplacians of an oriented
    simplicial complex.

    Parameters
    ----------
    S: simplicial complex object
        The simplicial complex of interest
    order: int, default: 1
        Specifies the order of the Hodge
        Laplacian matrix to be computed
    orientations: dict, default: None
        Dictionary mapping non-singleton simplices
        IDs to their boolean orientation
    index: bool, default: False
        Specifies whether to output dictionaries
        mapping the simplices IDs to indices

    Returns
    -------
    L_o: numpy.ndarray
        The Hodge Laplacian matrix of the chosen order, has dimension
        (n_simplices of given order, n_simplices of given order)
    matdict: dict
        The dictionary mapping indices to
        (order)-simplices IDs, if index is True

    """
    if index:
        B_o, __, matdict = boundary_matrix(S, order, orientations, True)
    else:
        B_o = boundary_matrix(S, order, orientations, False)
    D_om1 = np.transpose(B_o)

    B_op1 = boundary_matrix(S, order + 1, orientations, False)
    D_o = np.transpose(B_op1)

    L_o = D_om1 @ B_o + B_op1 @ D_o

    return (L_o, matdict) if index else L_o



"""Base class for directed hypergraphs.

.. warning::
    This is currently an experimental feature.

"""
from collections.abc import Hashable, Iterable
from copy import copy, deepcopy
from itertools import count
from warnings import warn

from ..exception import IDNotFound, XGIError, frozen
from ..utils import IDDict, update_uid_counter
from .diviews import DiEdgeView, DiNodeView

__all__ = ["DiHypergraph"]


class DiHypergraph:
    r"""A dihypergraph is a collection of directed interactions of arbitrary size.

    .. warning::
        This is currently an experimental feature.

    More formally, a directed hypergraph (dihypergraph) is a pair :math:`(V, E)`,
    where :math:`V` is a set of elements called *nodes* or *vertices*,
    and :math:`E` is the set of directed hyperedges.
    A directed hyperedge is an ordered pair, :math:`(e^+, e^-)`,
    where :math:`e^+ \subset V`, the set of senders, is known as the "tail" and
    :math:`e^-\subset V`, the set of receivers, is known as the "head".
    The equivalent undirected edge, is :math:`e = e^+ \cap e^-` and
    the edge size is defined as :math:`|e|`.

    The DiHypergraph class allows any hashable object as a node and can associate
    attributes to each node, edge, or the hypergraph itself, in the form of key/value
    pairs.

    Multiedges and self-loops are allowed.

    Parameters
    ----------
    incoming_data : input directed hypergraph data (optional, default: None)
        Data to initialize the dihypergraph. If None (default), an empty
        hypergraph is created, i.e. one with no nodes or edges.
        The data can be in the following formats:

        * directed hyperedge list
        * directed hyperedge dictionary
        * DiHypergraph object.

    **attr : dict, optional, default: None
        Attributes to add to the hypergraph as key, value pairs.

    Notes
    -----
    Unique IDs are assigned to each node and edge internally and are used to refer to
    them throughout.

    The `attr` keyword arguments are added as hypergraph attributes. To add node or edge
    attributes see :meth:`add_node` and :meth:`add_edge`.

    In addition to the methods listed in this page, other methods defined in the `stats`
    package are also accessible via the `DiHypergraph` class.  For more details, see the
    `tutorial
    <https://github.com/ComplexGroupInteractions/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

    References
    ----------
    Bretto, Alain. "Hypergraph theory: An introduction."
    Mathematical Engineering. Cham: Springer (2013).
    Examples
    --------
    >>> import xgi
    >>> DH = xgi.DiHypergraph([([1, 2, 3], [4]), ([5, 6], [6, 7, 8])])
    >>> DH.nodes
    DiNodeView((1, 2, 3, 4, 5, 6, 7, 8))
    >>> DH.edges
    DiEdgeView((0, 1))
    >>> [[sorted(h), sorted(t)] for h, t in DH.edges.dimembers()]
    [[[1, 2, 3], [4]], [[5, 6], [6, 7, 8]]]
    >>> [sorted(e) for e in DH.edges.members()]
    [[1, 2, 3, 4], [5, 6, 7, 8]]
    """
    _node_dict_factory = IDDict
    _node_attr_dict_factory = IDDict
    _hyperedge_dict_factory = IDDict
    _hyperedge_attr_dict_factory = IDDict
    _hypergraph_attr_dict_factory = dict

    def __getstate__(self):
        """Function that allows pickling.

        Returns
        -------
        dict
            The keys label the hypergraph dict and the values
            are dictionaries from the DiHypergraph class.

        Notes
        -----
        This allows the python multiprocessing module to be used.

        """
        return {
            "_edge_uid": self._edge_uid,
            "_hypergraph": self._hypergraph,
            "_node_in": self._node_in,
            "_node_out": self._node_out,
            "_node_attr": self._node_attr,
            "_edge_in": self._edge_in,
            "_edge_out": self._edge_out,
            "_edge_attr": self._edge_attr,
        }

    def __setstate__(self, state):
        """Function that allows unpickling of a dihypergraph.

        Parameters
        ----------
        state
            The keys access the dictionary names the values are the
            dictionarys themselves from the DiHypergraph class.

        Notes
        -----
        This allows the python multiprocessing module to be used.
        """
        self._edge_uid = state["_edge_uid"]
        self._hypergraph = state["_hypergraph"]
        self._node_in = state["_node_in"]
        self._node_out = state["_node_out"]
        self._node_attr = state["_node_attr"]
        self._edge_in = state["_edge_in"]
        self._edge_out = state["_edge_out"]
        self._edge_attr = state["_edge_attr"]
        self._nodeview = DiNodeView(self)
        self._edgeview = DiEdgeView(self)

    def __init__(self, incoming_data=None, **attr):
        self._edge_uid = count()
        self._hypergraph = self._hypergraph_attr_dict_factory()

        self._node_in = self._node_dict_factory()
        self._node_out = self._node_dict_factory()
        self._node_attr = self._node_attr_dict_factory()

        self._edge_in = self._hyperedge_dict_factory()
        self._edge_out = self._hyperedge_dict_factory()
        self._edge_attr = self._hyperedge_attr_dict_factory()

        self._nodeview = DiNodeView(self)
        """A :class:`~xgi.core.direportviews.DiNodeView` of the directed hypergraph."""

        self._edgeview = DiEdgeView(self)
        """An :class:`~xgi.core.direportviews.DiEdgeView` of the directed hypergraph."""

        if incoming_data is not None:
            # This import needs to happen when this function is called, not when it is
            # defined.  Otherwise, a circular import error would happen.
            from ..convert import convert_to_dihypergraph

            convert_to_dihypergraph(incoming_data, create_using=self)
        self._hypergraph.update(attr)  # must be after convert

    def __str__(self):
        """Returns a short summary of the directed hypergraph.

        Returns
        -------
        string
            DiHypergraph information

        """
        try:
            return f"{type(self).__name__} named {self['name']} with {self.num_nodes} nodes and {self.num_edges} hyperedges"
        except XGIError:
            return f"Unnamed {type(self).__name__} with {self.num_nodes} nodes and {self.num_edges} hyperedges"

    def __iter__(self):
        """Iterate over the nodes.

        Returns
        -------
        iterator
            An iterator over all nodes in the dihypergraph.
        """
        return iter(self._node_in)

    def __contains__(self, n):
        """Check for if a node is in this dihypergraph.

        Parameters
        ----------
        n : hashable
            node ID

        Returns
        -------
        bool
            Whether the node exists in the dihypergraph.
        """
        try:
            return n in self._node_in
        except TypeError:
            return False

    def __len__(self):
        """Number of nodes in the dihypergraph.

        Returns
        -------
        int
            The number of nodes in the dihypergraph.

        See Also
        --------
        num_nodes : identical method
        num_edges : number of edges in the dihypergraph

        """
        return len(self._node_in)

    def __getitem__(self, attr):
        """Read dihypergraph attribute."""
        try:
            return self._hypergraph[attr]
        except KeyError:
            raise XGIError("This attribute has not been set.")

    def __setitem__(self, attr, val):
        """Write dihypergraph attribute."""
        self._hypergraph[attr] = val

    def __getattr__(self, attr):
        stat = getattr(self.nodes, attr, None)
        word = "nodes"
        if stat is None:
            stat = getattr(self.edges, attr, None)
            word = "edges"
        if stat is None:
            word = None
            raise AttributeError(
                f"{attr} is not a method of DiHypergraph or a recognized DiNodeStat or DiEdgeStat"
            )

        def func(node=None, *args, **kwargs):
            val = stat(*args, **kwargs).asdict()
            return val if node is None else val[node]

        func.__doc__ = f"""Equivalent to DH.{word}.{attr}.asdict(). For accepted *args and
        **kwargs, see documentation of DH.{word}.{attr}."""

        return func

    @property
    def num_nodes(self):
        """The number of nodes in the dihypergraph.

        Returns
        -------
        int
            The number of nodes in the dihypergraph.

        See Also
        --------
        num_edges : returns the number of edges in the dihypergraph

        Examples
        --------
        >>> import xgi
        >>> hyperedge_list = [([1, 2], [2, 3, 4])]
        >>> DH = xgi.DiHypergraph(hyperedge_list)
        >>> DH.num_nodes
        4

        """
        return len(self._node_in)

    @property
    def num_edges(self):
        """The number of directed edges in the dihypergraph.

        Returns
        -------
        int
            The number of directed edges in the dihypergraph.

        See Also
        --------
        num_nodes : returns the number of nodes in the dihypergraph

        Examples
        --------
        >>> import xgi
        >>> hyperedge_list = [([1, 2], [2, 3, 4])]
        >>> DH = xgi.DiHypergraph(hyperedge_list)
        >>> DH.num_edges
        1
        """
        return len(self._edge_in)

    @property
    def nodes(self):
        """A :class:`DiNodeView` of this network."""
        return self._nodeview

    @property
    def edges(self):
        """An :class:`DiEdgeView` of this network."""
        return self._edgeview

    def add_node(self, node, **attr):
        """Add one node with optional attributes.

        Parameters
        ----------
        node : node
            A node can be any hashable Python object except None.
        attr : keyword arguments, optional
            Set or change node attributes using key=value.

        See Also
        --------
        add_nodes_from

        Notes
        -----
        If node is already in the dihypergraph, its attributes are still updated.

        """
        if node not in self._node_in:
            self._node_in[node] = set()
            self._node_out[node] = set()
            self._node_attr[node] = self._node_attr_dict_factory()
        self._node_attr[node].update(attr)

    def add_nodes_from(self, nodes_for_adding, **attr):
        """Add multiple nodes with optional attributes.

        Parameters
        ----------
        nodes_for_adding : iterable
            An iterable of nodes (list, dict, set, etc.).
            OR
            An iterable of (node, attribute dict) tuples.
            Node attributes are updated using the attribute dict.
        attr : keyword arguments, optional (default= no attributes)
            Update attributes for all nodes in nodes.
            Node attributes specified in nodes as a tuple take
            precedence over attributes specified via keyword arguments.

        See Also
        --------
        add_node

        """
        for n in nodes_for_adding:
            try:
                newnode = n not in self._node_in
                newdict = attr
            except TypeError:
                n, ndict = n
                newnode = n not in self._node_in
                newdict = attr.copy()
                newdict.update(ndict)
            if newnode:
                self._node_in[n] = set()
                self._node_out[n] = set()
                self._node_attr[n] = self._node_attr_dict_factory()
            self._node_attr[n].update(newdict)

    def remove_node(self, n, strong=False):
        """Remove a single node.

        The removal may be weak (default) or strong.  In weak removal, the node is
        removed from each of its containing edges.  If it is contained in any singleton
        edges, then these are also removed.  In strong removal, all edges containing the
        node are removed, regardless of size.

        Parameters
        ----------
        n : node
            A node in the dihypergraph

        strong : bool (default False)
            Whether to execute weak or strong removal.

        Raises
        ------
        XGIError
           If n is not in the dihypergraph.

        See Also
        --------
        remove_nodes_from

        """
        out_edge_neighbors = self._node_in[n]
        in_edge_neighbors = self._node_out[n]
        del self._node_in[n]
        del self._node_out[n]
        del self._node_attr[n]

        if strong:
            for edge in in_edge_neighbors.union(out_edge_neighbors):
                del self._edge_in[edge]
                del self._edge_out[edge]
                del self._edge_attr[edge]
        else:  # weak removal
            for edge in in_edge_neighbors:
                self._edge_in[edge].remove(n)

            for edge in out_edge_neighbors:
                self._edge_out[edge].remove(n)

            # remove empty edges
            for edge in in_edge_neighbors.union(out_edge_neighbors):
                if not self._edge_in[edge] and not self._edge_out[edge]:
                    del self._edge_in[edge]
                    del self._edge_out[edge]
                    del self._edge_attr[edge]

    def remove_nodes_from(self, nodes):
        """Remove multiple nodes.

        Parameters
        ----------
        nodes : iterable
            An iterable of nodes.

        See Also
        --------
        remove_node

        """
        for n in nodes:
            if n not in self._node_in:
                warn(f"Node {n} not in dihypergraph")
                continue
            self.remove_node(n)

    def set_node_attributes(self, values, name=None):
        """Sets node attributes from a given value or dictionary of values.

        Parameters
        ----------
        values : scalar value, dict-like
            What the node attribute should be set to.  If `values` is
            not a dictionary, then it is treated as a single attribute value
            that is then applied to every node in `DH`.  This means that if
            you provide a mutable object, like a list, updates to that object
            will be reflected in the node attribute for every node.
            The attribute name will be `name`.

            If `values` is a dict or a dict of dict, it should be keyed
            by node to either an attribute value or a dict of attribute key/value
            pairs used to update the node's attributes.
        name : string, optional
            Name of the node attribute to set if values is a scalar, by default None.

        See Also
        --------
        set_edge_attributes
        add_node
        add_nodes_from

        Notes
        -----
        After computing some property of the nodes of a hypergraph, you may
        want to assign a node attribute to store the value of that property
        for each node.

        If you provide a list as the second argument, updates to the list
        will be reflected in the node attribute for each node.

        If you provide a dictionary of dictionaries as the second argument,
        the outer dictionary is assumed to be keyed by node to an inner
        dictionary of node attributes for that node.

        Note that if the dictionary contains nodes that are not in `G`, the
        values are silently ignored.

        """
        # Set node attributes based on type of `values`
        if name is not None:  # `values` must not be a dict of dict
            if isinstance(values, dict):  # `values` is a dict
                for n, v in values.items():
                    try:
                        self._node_attr[n][name] = v
                    except IDNotFound:
                        warn(f"Node {n} does not exist!")
            else:  # `values` is a constant
                for n in self:
                    self._node_attr[n][name] = values
        else:  # `values` must be dict of dict
            try:
                for n, d in values.items():
                    try:
                        self._node_attr[n].update(d)
                    except IDNotFound:
                        warn(f"Node {n} does not exist!")
            except (TypeError, ValueError, AttributeError):
                raise XGIError("Must pass a dictionary of dictionaries")

    def add_edge(self, members, id=None, **attr):
        """Add one edge with optional attributes.

        Parameters
        ----------
        members : Iterable
            An list or tuple (size 2) of iterables. The first entry contains the
            elements of the tail and the second entry contains the elements
            of the head.
        id : hashable, default None
            Id of the new edge. If None, a unique numeric ID will be created.
        **attr : dict, optional
            Attributes of the new edge.

        Raises
        -----
        XGIError
            If `members` is empty or is not a list or tuple.

        See Also
        --------
        add_edges_from : Add a collection of edges.

        Examples
        --------

        Add edges with or without specifying an edge id.

        >>> import xgi
        >>> DH = xgi.DiHypergraph()
        >>> DH.add_edge(([1, 2, 3], [2, 3, 4]))
        >>> DH.add_edge(([3, 4], set()), id='myedge')
        """
        if not members:
            raise XGIError("Cannot add an empty edge")

        if isinstance(members, (tuple, list)):
            tail = members[0]
            head = members[1]
        else:
            raise XGIError("Directed edge must be a list or tuple!")

        if not head and not tail:
            raise XGIError("Cannot add an empty edge")

        uid = next(self._edge_uid) if id is None else id

        if id in self._edge_in.keys():  # check that uid is not present yet
            warn(f"uid {id} already exists, cannot add edge {members}")
            return

        self._edge_in[uid] = set()
        self._edge_out[uid] = set()

        for node in tail:
            if node not in self._node_in:
                self._node_in[node] = set()
                self._node_out[node] = set()
                self._node_attr[node] = self._node_attr_dict_factory()
            self._node_in[node].add(uid)
            self._edge_out[uid].add(node)

        for node in head:
            if node not in self._node_out:
                self._node_in[node] = set()
                self._node_out[node] = set()
                self._node_attr[node] = self._node_attr_dict_factory()
            self._node_out[node].add(uid)
            self._edge_in[uid].add(node)

        self._edge_attr[uid] = self._hyperedge_attr_dict_factory()
        self._edge_attr[uid].update(attr)

        if id:  # set self._edge_uid correctly
            update_uid_counter(self, id)

    def add_edges_from(self, ebunch_to_add, **attr):
        """Add multiple directed edges with optional attributes.

        Parameters
        ----------
        ebunch_to_add : Iterable

            Note that here, when we refer to an edge, as in the `add_edge` method,
            it is a list or tuple (size 2) of iterables. The first entry contains the
            elements of the tail and the second entry contains the elements
            of the head.

            An iterable of edges.  This may be an iterable of edges (Format 1),
            where each edge is in the format described above.

            Alternatively, each element could also be a tuple in any of the following
            formats:

            * Format 2: 2-tuple (edge, edge_id), or
            * Format 4: 3-tuple (edge, edge_id, attr),

            where `edge` is in the format described above, `edge_id` is a hashable to use
            as edge ID, and `attr` is a dict of attributes. Finally, `ebunch_to_add`
            may be a dict of the form `{edge_id: edge_members}` (Format 5).

            Formats 2 and 3 are unambiguous because `attr` dicts are not hashable, while `id`s must be.
            In Formats 2-4, each element of `ebunch_to_add` must have the same length,
            i.e. you cannot mix different formats.  The iterables containing edge
            members cannot be strings.

        attr : \*\*kwargs, optional
            Additional attributes to be assigned to all edges. Attribues specified via
            `ebunch_to_add` take precedence over `attr`.

        See Also
        --------
        add_edge : Add a single edge.

        Notes
        -----
        Adding the same edge twice will create a multi-edge. Currently
        cannot add empty edges; the method skips over them.

        Examples
        --------
        >>> import xgi
        >>> DH = xgi.DiHypergraph()

        When specifying edges by their members only, numeric edge IDs will be assigned
        automatically.

        >>> DH.add_edges_from([([0, 1], [1, 2]), ([2, 3, 4], [])])
        >>> DH.edges.dimembers(dtype=dict)
        {0: ({0, 1}, {1, 2}), 1: ({2, 3, 4}, set())}

        Custom edge ids can be specified using a dict.

        >>> DH = xgi.DiHypergraph()
        >>> DH.add_edges_from({'one': ([0, 1], [1, 2]), 'two': ([2, 3, 4], [])})
        >>> DH.edges.dimembers(dtype=dict)
        {'one': ({0, 1}, {1, 2}), 'two': ({2, 3, 4}, set())}

        You can use the dict format to easily add edges from another hypergraph.

        >>> DH2 = xgi.DiHypergraph()
        >>> DH2.add_edges_from(DH.edges.dimembers(dtype=dict))
        >>> DH.edges == DH2.edges
        True

        Alternatively, edge ids can be specified using an iterable of 2-tuples.

        >>> DH = xgi.DiHypergraph()
        >>> DH.add_edges_from([(([0, 1], [1, 2]), 'one'), (([2, 3, 4], []), 'two')])
        >>> DH.edges.dimembers(dtype=dict)
        {'one': ({0, 1}, {1, 2}), 'two': ({2, 3, 4}, set())}

        Attributes for each edge may be specified using a 2-tuple for each edge.
        Numeric IDs will be assigned automatically.

        >>> DH = xgi.DiHypergraph()
        >>> edges = [
        ...     (([0, 1], [1, 2]), {'color': 'red'}),
        ...     (([2, 3, 4], []), {'color': 'blue', 'age': 40}),
        ... ]
        >>> DH.add_edges_from(edges)
        >>> {e: DH.edges[e] for e in DH.edges}
        {0: {'color': 'red'}, 1: {'color': 'blue', 'age': 40}}

        Attributes and custom IDs may be specified using a 3-tuple for each edge.

        >>> DH = xgi.DiHypergraph()
        >>> edges = [
        ...     (([0, 1], [1, 2]), 'one', {'color': 'red'}),
        ...     (([2, 3, 4], []), 'two', {'color': 'blue', 'age': 40}),
        ... ]
        >>> DH.add_edges_from(edges)
        >>> {e: DH.edges[e] for e in DH.edges}
        {'one': {'color': 'red'}, 'two': {'color': 'blue', 'age': 40}}

        """
        # format 5 is the easiest one
        if isinstance(ebunch_to_add, dict):
            for id, members in ebunch_to_add.items():
                if id in self._edge_in.keys():  # check that uid is not present yet
                    warn(f"uid {id} already exists, cannot add edge {members}.")
                    continue

                if isinstance(members, (tuple, list)):
                    tail = members[0]
                    head = members[1]
                else:
                    raise XGIError("Directed edge must be a list or tuple!")

                try:
                    self._edge_in[id] = set(head)
                    self._edge_out[id] = set(tail)
                except TypeError as e:
                    raise XGIError("Invalid ebunch format") from e

                for n in tail:
                    if n not in self._node_in:
                        self._node_in[n] = set()
                        self._node_out[n] = set()
                        self._node_attr[n] = self._node_attr_dict_factory()
                    self._node_in[n].add(id)
                self._edge_attr[id] = self._hyperedge_attr_dict_factory()

                for n in head:
                    if n not in self._node_in:
                        self._node_in[n] = set()
                        self._node_out[n] = set()
                        self._node_attr[n] = self._node_attr_dict_factory()
                    self._node_out[n].add(id)

                update_uid_counter(self, id)

            return
        # in formats 1-4 we only know that ebunch_to_add is an iterable, so we iterate
        # over it and use the firs element to determine which format we are working with
        new_edges = iter(ebunch_to_add)
        try:
            first_edge = next(new_edges)
        except StopIteration:
            return

        second_elem = list(first_edge)[1]

        format1, format2, format3, format4 = False, False, False, False

        if (
            isinstance(second_elem, Iterable)
            and not isinstance(second_elem, str)
            and not isinstance(second_elem, dict)
        ):
            format1 = True
        else:
            if len(first_edge) == 3:
                format4 = True
            elif len(first_edge) == 2 and issubclass(type(first_edge[1]), Hashable):
                format2 = True
            elif len(first_edge) == 2:
                format3 = True

        # now we may iterate over the rest
        e = first_edge
        while True:
            if format1:
                members, id, eattr = e, next(self._edge_uid), {}
            elif format2:
                members, id, eattr = e[0], e[1], {}
            elif format3:
                members, id, eattr = e[0], next(self._edge_uid), e[1]
            elif format4:
                members, id, eattr = e[0], e[1], e[2]

            if id in self._edge_in.keys():  # check that uid is not present yet
                warn(f"uid {id} already exists, cannot add edge {members}.")
            else:
                try:
                    tail = members[0]
                    head = members[1]
                    self._edge_out[id] = set(tail)
                    self._edge_in[id] = set(head)
                except TypeError as e:
                    raise XGIError("Invalid ebunch format") from e

                for node in tail:
                    if node not in self._node_in:
                        self._node_in[node] = set()
                        self._node_out[node] = set()
                        self._node_attr[node] = self._node_attr_dict_factory()
                    self._node_in[node].add(id)
                    self._edge_out[id].add(node)

                for node in head:
                    if node not in self._node_out:
                        self._node_in[node] = set()
                        self._node_out[node] = set()
                        self._node_attr[node] = self._node_attr_dict_factory()
                    self._node_out[node].add(id)
                    self._edge_in[id].add(node)

                self._edge_attr[id] = self._hyperedge_attr_dict_factory()
                self._edge_attr[id].update(attr)
                self._edge_attr[id].update(eattr)

            try:
                e = next(new_edges)
            except StopIteration:
                if format2 or format4:
                    update_uid_counter(self, id)
                break

    def remove_edge(self, id):
        """Remove one edge.

        Parameters
        ----------
        id : Hashable
            edge ID to remove

        Raises
        ------
        XGIError
            If no edge has that ID.

        See Also
        --------
        remove_edges_from : Remove multiple edges.

        """
        head = self._edge_in[id].copy()
        tail = self._edge_out[id].copy()

        for node in head:
            self._node_out[node].remove(id)
        for node in tail:
            self._node_in[node].remove(id)

        del self._edge_in[id]
        del self._edge_out[id]
        del self._edge_attr[id]

    def remove_edges_from(self, ebunch):
        """Remove multiple edges.

        Parameters
        ----------
        ebunch: Iterable
            Edges to remove.

        Raises
        ------
        xgi.exception.IDNotFound
            If an id in ebunch is not part of the network.

        See Also
        --------
        remove_edge : remove a single edge.

        """
        for id in ebunch:
            head = self._edge_in[id].copy()
            tail = self._edge_out[id].copy()

            for node in head:
                self._node_out[node].remove(id)
            for node in tail:
                self._node_in[node].remove(id)

            del self._edge_in[id]
            del self._edge_out[id]
            del self._edge_attr[id]

    def set_edge_attributes(self, values, name=None):
        """Set the edge attributes from a value or a dictionary of values.

        Parameters
        ----------
        values : scalar value, dict-like
            What the edge attribute should be set to.  If `values` is
            not a dictionary, then it is treated as a single attribute value
            that is then applied to every edge in `DH`.  This means that if
            you provide a mutable object, like a list, updates to that object
            will be reflected in the edge attribute for each edge.  The attribute
            name will be `name`.
            If `values` is a dict or a dict of dict, it should be keyed
            by edge ID to either an attribute value or a dict of attribute
            key/value pairs used to update the edge's attributes.
        name : string, optional
            Name of the edge attribute to set if values is a scalar. By default, None.

        See Also
        --------
        set_node_attributes
        add_edge
        add_edges_from

        Notes
        -----
        Note that if the dict contains edge IDs that are not in `DH`, they are
        silently ignored.

        """
        if name is not None:
            # `values` does not contain attribute names
            try:
                for e, value in values.items():
                    try:
                        self._edge_attr[id][name] = value
                    except IDNotFound:
                        warn(f"Edge {e} does not exist!")
            except AttributeError:
                # treat `values` as a constant
                for e in self._edge_in:
                    self._edge_attr[e][name] = values
        else:
            try:
                for e, d in values.items():
                    try:
                        self._edge_attr[e].update(d)
                    except IDNotFound:
                        warn(f"Edge {e} does not exist!")
            except AttributeError:
                raise XGIError(
                    "name property has not been set and a "
                    "dict-of-dicts has not been provided."
                )

    def clear(self, hypergraph_attr=True):
        """Remove all nodes and edges from the graph.

        Also removes node and edge attributes, and optionally hypergraph attributes.

        Parameters
        ----------
        hypergraph_attr : bool, optional
            Whether to remove hypergraph attributes as well.
            By default, True.

        """
        self._node_in.clear()
        self._node_out.clear()
        self._node_attr.clear()
        self._edge_in.clear()
        self._edge_out.clear()
        self._edge_attr.clear()
        if hypergraph_attr:
            self._hypergraph.clear()

    def copy(self):
        """A deep copy of the dihypergraph.

        A deep copy of the dihypergraph, including node, edge, and hypergraph attributes.

        Returns
        -------
        DH : DiHypergraph
            A copy of the hypergraph.

        """
        cp = self.__class__()
        nn = self.nodes
        cp.add_nodes_from((n, deepcopy(attr)) for n, attr in nn.items())
        ee = self.edges
        cp.add_edges_from(
            (e, id, deepcopy(self.edges[id]))
            for id, e in ee.dimembers(dtype=dict).items()
        )
        cp._hypergraph = deepcopy(self._hypergraph)

        cp._edge_uid = copy(self._edge_uid)

        return cp

    def cleanup(self, isolates=False, relabel=True, in_place=True):
        if in_place:
            if not isolates:
                self.remove_nodes_from(self.nodes.isolates())
            if relabel:
                from ..utils import convert_labels_to_integers

                temp = convert_labels_to_integers(self).copy()

                nn = temp.nodes
                ee = temp.edges

                self.clear()
                self.add_nodes_from((n, deepcopy(attr)) for n, attr in nn.items())
                self.add_edges_from(
                    (e, id, deepcopy(temp.edges[id]))
                    for id, e in ee.dimembers(dtype=dict).items()
                )
                self._hypergraph = deepcopy(temp._hypergraph)
        else:
            DH = self.copy()
            if not isolates:
                DH.remove_nodes_from(DH.nodes.isolates())
            if relabel:
                from ..utils import convert_labels_to_integers

                DH = convert_labels_to_integers(DH)

            return DH

    def freeze(self):
        """Method for freezing a dihypergraph which prevents it from being modified

        See Also
        --------
        frozen : Method that raises an error when a user tries to modify the hypergraph
        is_frozen : Check whether a dihypergraph is frozen

        Examples
        --------
        >>> import xgi
        >>> diedgelist = [([1, 2], [2, 3, 4])]
        >>> DH = xgi.DiHypergraph(diedgelist)
        >>> DH.freeze()
        >>> DH.add_node(5)
        Traceback (most recent call last):
        xgi.exception.XGIError: Frozen higher-order network can't be modified

        """
        self.add_node = frozen
        self.add_nodes_from = frozen
        self.remove_node = frozen
        self.remove_nodes_from = frozen
        self.add_edge = frozen
        self.add_edges_from = frozen
        self.add_weighted_edges_from = frozen
        self.remove_edge = frozen
        self.remove_edges_from = frozen
        self.add_node_to_edge = frozen
        self.remove_node_from_edge = frozen
        self.clear = frozen
        self.frozen = True

    @property
    def is_frozen(self):
        """Checks whether a hypergraph is frozen

        Returns
        -------
        bool
            True if hypergraph is frozen, false if not.

        See Also
        --------
        freeze : A method to prevent a hypergraph from being modified.

        Examples
        --------
        >>> import xgi
        >>> edges = [([1, 2], [2, 3, 4])]
        >>> DH = xgi.DiHypergraph(edges)
        >>> DH.freeze()
        >>> DH.is_frozen
        True

        """
        try:
            return self.frozen
        except AttributeError:
            return False



"""Functions to randomize hypergaphs

All the functions in this module return a Hypergraph class (i.e. a simple, undirected
hypergraph).

"""

import random
from copy import deepcopy

import xgi

__all__ = [
    "shuffle_hyperedges",
    "node_swap",
]


def shuffle_hyperedges(S, order, p):
    """Shuffle existing hyperdeges of order `order` with probablity `p`.

    Parameters
    ----------
    S : xgi.HyperGraph
        Hypergraph
    order : int
        Order of hyperedges to shuffle
    p : float
        Probability of shuffling each hyperedge

    Returns
    -------
    H: xgi.HyperGraph
        Hypergraph with edges of order d shuffled

    Note
    ----
    By shuffling hyperedges in a simplicial complex, it will in general lose
    its "simpliciality" and become a hypergraph.

    Reference
    ---------
    Zhang, Y.*, Lucas, M.* and Battiston, F., 2023.
    "Higher-order interactions shape collective dynamics differently
    in hypergraphs and simplicial complexes."
    Nature Communications, 14(1), p.1605.
    https://doi.org/10.1038/s41467-023-37190-9

    Example
    -------
    >>> S = xgi.random_simplicial_complex(50, [0.1, 0.01, 0.001], seed=1)
    >>> H = xgi.shuffle_hyperedges(S, order=2, p=0.5)

    """

    if (order + 1) not in xgi.unique_edge_sizes(S):
        raise ValueError(f"There is no hyperedge of order {order} is this hypergraph.")
    if (p < 0) or (p > 1):
        raise ValueError("p must be between 0 and 1 included.")

    # convert to Hypergraph to be able to shuffle edges
    if isinstance(S, xgi.Hypergraph):
        H = xgi.Hypergraph()
        H.add_nodes_from(S.nodes)
        H.add_edges_from(S._edge)
    else:
        H = S.copy()

    nodes = list(S.nodes)
    d_hyperedges = H.edges.filterby("order", order).members(dtype=dict)

    for id_, members in d_hyperedges.items():
        if random.random() <= p:
            H.remove_edge(id_)
            new_hyperedge = tuple(random.sample(nodes, order + 1))
            while new_hyperedge in H._edge.values():
                new_hyperedge = tuple(random.sample(nodes, order + 1))
            H.add_edge(new_hyperedge)

    return H


def node_swap(H, nid1, nid2, id_temp=-1, order=None):
    """Swap nodes `nid1` and node `nid2` in all edges of order `order`.

    Parameters
    ----------
    H: HyperGraph
        Hypergraph to consider
    nid1: node ID
        ID of first node to swap
    nid2: node ID
        ID of second node to swap
    id_temp: node ID
        Temporary ID given to nodes when swapping
    order: {int, None}, default: None
        If None, consider all orders. If an integer,
        consider edges of that order.

    Returns
    -------
    HH: HyperGraph

    Reference
    ---------
    Zhang, Y.*, Lucas, M.* and Battiston, F., 2023.
    "Higher-order interactions shape collective dynamics differently
    in hypergraphs and simplicial complexes."
    Nature Communications, 14(1), p.1605.
    https://doi.org/10.1038/s41467-023-37190-9

    """

    # check that node ids are in hypergraph
    if not nid1 in H:
        raise ValueError(f"Node {nid1} is not in hypergraph H")
    if not nid2 in H:
        raise ValueError(f"Node {nid2} is not in hypergraph H")

    if order is not None:
        if (order + 1) not in xgi.unique_edge_sizes(H):
            raise ValueError(
                f"There is no hyperedge of order {order} is this hypergraph."
            )

    # make sure id_temps does not exist yet
    while id_temp in H.edges:
        id_temp -= 1

    # get edges of given order
    if order:
        edge_dict = H.edges.filterby("order", order).members(dtype=dict).copy()
    else:  # includes order 0
        edge_dict = H.edges.members(dtype=dict).copy()

    # check that node ids exist in those edges
    if H.nodes.degree(order=order)[nid1] == 0:
        raise ValueError(
            f"Node {nid1} is not part of any hyperedge of the specified order"
        )
    if H.nodes.degree(order=order)[nid2] == 0:
        raise ValueError(
            f"Node {nid2} is not part of any hyperedge of the specified order"
        )

    new_edge_dict = deepcopy(edge_dict)
    HH = H.copy()

    # replace nid1 by temporary id in edges
    for key, members in edge_dict.items():
        if nid1 in members:
            members.remove(nid1)
            members.add(id_temp)
        new_edge_dict[key] = members

    # replace nid2 by nid1 in edges
    for key, members in new_edge_dict.items():
        if nid2 in members:
            members.remove(nid2)
            members.add(nid1)
        new_edge_dict[key] = members

    # replace temporary id by nid2 in edges
    for key, members in new_edge_dict.items():
        if id_temp in members:
            members.remove(id_temp)
            members.add(nid2)
        new_edge_dict[key] = members

    # update hypergraph with new edges
    HH.remove_edges_from(edge_dict)
    HH.add_edges_from(new_edge_dict)

    return HH
